# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, James Diprose


from __future__ import annotations

import gzip
import logging
import os
import shutil
from typing import Dict, List
from zipfile import ZipFile

import pendulum
import requests
from airflow.models.taskinstance import TaskInstance
from google.cloud.bigquery import SourceFormat

from academic_observatory_workflows.config import schema_folder as default_schema_folder
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.data_utils import get_file
from observatory.platform.utils.workflow_utils import upload_files_from_list
from observatory.platform.workflows.snapshot_telescope import (
    SnapshotRelease,
    SnapshotTelescope,
)


def fetch_release_date() -> pendulum.DateTime:
    """Fetch the Geonames release date.

    :return: the release date.
    """

    response = requests.head(GeonamesRelease.DOWNLOAD_URL)
    date_str = response.headers["Last-Modified"]
    date: pendulum.DateTime = pendulum.from_format(date_str, "ddd, DD MMM YYYY HH:mm:ss z")
    return date


def first_sunday_of_month(datetime: pendulum.DateTime) -> pendulum.DateTime:
    """Get the first Sunday of the month based on a given datetime.

    :param datetime: the datetime.
    :return: the first Sunday of the month.
    """

    return datetime.start_of("month").first_of("month", day_of_week=7)


class GeonamesRelease(SnapshotRelease):
    DOWNLOAD_URL = "https://download.geonames.org/export/dump/allCountries.zip"

    def __init__(self, dag_id: str, release_date: pendulum.DateTime):
        """Create a GeonamesRelease instance.

        :param dag_id: the DAG id.
        :param release_date: the date of the release.
        """

        download_file_name = f"{dag_id}.zip"
        extract_file_name = f"allCountries.txt"
        transform_file_name = f"{dag_id}.csv.gz"
        super().__init__(dag_id, release_date, download_file_name, extract_file_name, transform_file_name)

    @property
    def download_path(self) -> str:
        """Get the path to the downloaded file.

        :return: the file path.
        """

        return os.path.join(self.download_folder, self.download_files_regex)

    @property
    def extract_path(self) -> str:
        """Get the path to the extracted file.

        :return: the file path.
        """

        return os.path.join(self.extract_folder, self.extract_files_regex)

    @property
    def transform_path(self) -> str:
        """Get the path to the transformed file.

        :return: the file path.
        """

        return os.path.join(self.transform_folder, self.transform_files_regex)

    def download(self):
        """Downloads geonames dump file containing country data. The file is in zip format and will be extracted
        after downloading, saving the unzipped content.

        :return: None
        """

        file_path, updated = get_file(
            fname=self.download_path, origin=GeonamesRelease.DOWNLOAD_URL, cache_subdir="", extract=False
        )

        logging.info(f"Downloaded file: {file_path}")

    def extract(self):
        """Extract a downloaded Geonames release.

        :return: None
        """

        with ZipFile(self.download_path) as zip_file:
            zip_file.extractall(self.extract_folder)

    def transform(self):
        """Transforms release by storing file content in gzipped csv format.

        :return: None
        """

        with open(self.extract_path, "rb") as file_in:
            with gzip.open(self.transform_path, "wb") as file_out:
                shutil.copyfileobj(file_in, file_out)


class GeonamesTelescope(SnapshotTelescope):
    """
    A Telescope that harvests the GeoNames geographical database: https://www.geonames.org/

    Saved to the BigQuery table: <project_id>.geonames.geonamesYYYYMMDD
    """

    DAG_ID = "geonames"

    def __init__(
        self,
        dag_id: str = DAG_ID,
        start_date: pendulum.DateTime = pendulum.datetime(2020, 9, 1),
        schedule_interval: str = "@weekly",
        dataset_id: str = "geonames",
        schema_folder: str = default_schema_folder(),
        source_format: str = SourceFormat.CSV,
        dataset_description: str = "The GeoNames geographical database: https://www.geonames.org/",
        load_bigquery_table_kwargs: Dict = None,
        table_descriptions: Dict = None,
        catchup: bool = False,
        airflow_vars: List = None,
    ):

        """The Geonames telescope.

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param dataset_id: the BigQuery dataset id.
        :param schema_folder: the SQL schema path.
        :param source_format: the format of the data to load into BigQuery.
        :param dataset_description: description for the BigQuery dataset.
        :param load_bigquery_table_kwargs: the customisation parameters for loading data into a BigQuery table.
        :param table_descriptions: a dictionary with table ids and corresponding table descriptions.
        :param catchup:  whether to catchup the DAG or not.
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow.
        """

        if load_bigquery_table_kwargs is None:
            load_bigquery_table_kwargs = {"csv_field_delimiter": "\t", "csv_quote_character": ""}

        if table_descriptions is None:
            table_descriptions = {dag_id: "The GeoNames table."}

        if airflow_vars is None:
            airflow_vars = [
                AirflowVars.DATA_PATH,
                AirflowVars.PROJECT_ID,
                AirflowVars.DATA_LOCATION,
                AirflowVars.DOWNLOAD_BUCKET,
                AirflowVars.TRANSFORM_BUCKET,
            ]
        super().__init__(
            dag_id,
            start_date,
            schedule_interval,
            dataset_id,
            schema_folder,
            source_format=source_format,
            load_bigquery_table_kwargs=load_bigquery_table_kwargs,
            dataset_description=dataset_description,
            table_descriptions=table_descriptions,
            catchup=catchup,
            airflow_vars=airflow_vars,
        )
        self.add_setup_task(self.check_dependencies)
        self.add_setup_task(self.fetch_release_date)
        self.add_task(self.download)
        self.add_task(self.upload_downloaded)
        self.add_task(self.extract)
        self.add_task(self.transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> List[GeonamesRelease]:
        """Make release instances. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are
        passed to this argument.
        :return: a list of GeonamesRelease instances.
        """

        ti: TaskInstance = kwargs["ti"]
        release_date = ti.xcom_pull(
            key=GeonamesTelescope.RELEASE_INFO, task_ids=self.fetch_release_date.__name__, include_prior_dates=False
        )

        return [GeonamesRelease(self.dag_id, pendulum.parse(release_date))]

    def fetch_release_date(self, **kwargs):
        """Get the Geonames release for a given month and publishes the release_date as an XCom.

        :param kwargs: the context passed from the BranchPythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: whether to keep executing the DAG.
        """

        # Check if first Sunday of month
        execution_date = kwargs["execution_date"]
        run_date = first_sunday_of_month(execution_date)
        logging.info(f"execution_date={execution_date}, run_date={run_date}")

        # If first Sunday of month get current release date and push for processing
        continue_dag = execution_date == run_date
        if continue_dag:
            # Fetch release date
            release_date = fetch_release_date()

            # Push messages
            ti: TaskInstance = kwargs["ti"]
            ti.xcom_push(GeonamesTelescope.RELEASE_INFO, release_date.format("YYYYMMDD"), execution_date)

        return continue_dag

    def download(self, releases: List[GeonamesRelease], **kwargs):
        """Task to download the GeonamesRelease release for a given month.

        :param releases: the list of GeonamesRelease instances.
        :return: None.
        """

        # Download each release
        for release in releases:
            release.download()

    def upload_downloaded(self, releases: List[GeonamesRelease], **kwargs):
        """Task to upload the downloaded GeonamesRelease release for a given month.

        :param releases: the list of GeonamesRelease instances.
        :return: None.
        """

        for release in releases:
            upload_files_from_list(release.download_files, release.download_bucket)

    def extract(self, releases: List[GeonamesRelease], **kwargs):
        """Task to extract the GeonamesRelease release for a given month.

        :param release: GeonamesRelease.
        :return: None.
        """

        for release in releases:
            release.extract()

    def transform(self, releases: List[GeonamesRelease], **kwargs):
        """Task to transform the GeonamesRelease release for a given month.

        :param releases: the list of GeonamesRelease instances.
        :return: None.
        """

        for release in releases:
            release.transform()



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien


import json
import logging
import os
import urllib.request
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed
from math import ceil
from typing import Any, Dict, List, Type, Union
from urllib.error import URLError

import backoff
import pendulum
import xmltodict
from airflow.exceptions import AirflowException
from airflow.models import Variable
from airflow.models.taskinstance import TaskInstance
from google.cloud.bigquery import SourceFormat, WriteDisposition
from ratelimit import limits, sleep_and_retry
from suds import WebFault
from wos import WosClient

from academic_observatory_workflows.config import schema_folder
from observatory.platform.utils.airflow_utils import AirflowVars, check_variables
from observatory.platform.utils.config_utils import find_schema
from observatory.platform.utils.file_utils import write_to_file, zip_files
from observatory.platform.utils.gc_utils import (
    bigquery_sharded_table_id,
    create_bigquery_dataset,
    load_bigquery_table,
)
from observatory.platform.utils.workflow_utils import SubFolder, workflow_path
from observatory.platform.utils.workflow_utils import (
    build_schedule,
    delete_msg_files,
    get_as_list,
    get_as_list_or_none,
    get_entry_or_none,
    json_to_db,
    upload_telescope_file_list,
    validate_date,
    write_xml_to_json,
)


class WosUtilConst:
    """Class containing some WosUtility constants. Makes these values accessible by decorators."""

    # WoS limits as a guide for reference. Throttle limits more conservative than this.
    RESULT_LIMIT = 100  # Return 100 results max per query.
    CALL_LIMIT = 1  # WoS says they can do 2 api calls / second, but we're setting 1 per 2 seconds.
    CALL_PERIOD = 2  # seconds
    SESSION_CALL_LIMIT = 5  # 5 calls.
    SESSION_CALL_PERIOD = 360  # 6 minutes. [Actual WoS limit is 5 mins]
    RETRIES = 3


class WosUtility:
    """Handles the interaction with Web of Science"""

    @staticmethod
    def build_query(wos_inst_id: List[str], period: Type[pendulum.Period]) -> OrderedDict:
        """Build a WoS API query.

        :param wos_inst_id: List of Institutional ID to query, e.g, "Curtin University"
        :param period: A tuple containing start and end dates.
        :return: Constructed web query.
        """
        start_date = period.start.isoformat()
        end_date = period.end.isoformat()

        organisations = str()
        n_institutes = len(wos_inst_id)
        for i, inst in enumerate(wos_inst_id):
            organisations += inst
            organisations += " OR "

        organisations = organisations[:-4]  # Remove last ' OR '

        query_str = f"OG=({organisations})"

        logging.info(f"Query string: {query_str}")
        query = OrderedDict(
            [
                ("query", query_str),
                ("count", WosUtilConst.RESULT_LIMIT),
                ("offset", 1),
                ("timeSpan", {"begin": start_date, "end": end_date}),
            ]
        )
        return query

    @staticmethod
    def parse_query(records: Any) -> (dict, str):
        """Parse XML tree record into a dict.

        :param records: XML tree returned by the web query.
        :return: Dictionary version of the web response and a schema version string.
        """

        try:
            records_dict = xmltodict.parse(records)
        except Exception as e:
            raise AirflowException(f"XML to dict parsing failed: {e}")

        records = records_dict["records"]
        schema_string = records["@xmlns"]

        prefix_len = len(WosTelescope.SCHEMA_ID_PREFIX)
        prefix_loc = schema_string.find(WosTelescope.SCHEMA_ID_PREFIX)
        if prefix_loc == -1:
            logging.warning(
                f"WOS schema has changed.\nExpecting prefix: {WosTelescope.SCHEMA_ID_PREFIX}\n"
                f"Received: {schema_string}"
            )
            return None

        ver_start = prefix_len
        ver_end = schema_string.find("/", ver_start)
        schema_ver = schema_string[ver_start:ver_end]
        logging.info(f"Found schema version: {schema_ver}")

        if "REC" not in records:
            logging.warning(f"No record found for query. Please double check parameters, e.g., institution id.")
            return None, schema_ver

        return get_as_list(records, "REC"), schema_ver

    @staticmethod
    def download_wos_period(
        client: WosClient, conn: str, period: pendulum.Period, wos_inst_id: List[str], download_path: str
    ) -> List[str]:
        """Download records for a stated date range.

        :param client: WebClient object.
        :param conn: file name for saved response as a pickle file.
        :param period: Period tuple containing (start date, end date).
        :param wos_inst_id: List of Institutional ID to query, e.g, "Curtin University"
        :param download_path: Path to download files to.
        """

        timestamp = pendulum.now().isoformat()
        inst_str = conn[WosTelescope.ID_STRING_OFFSET :]
        save_file_prefix = os.path.join(
            download_path, period.start.isoformat(), inst_str, f"{period.start}-{period.end}_{timestamp}"
        )
        query = WosUtility.build_query(wos_inst_id, period)
        result = WosUtility.make_query(client, query)
        logging.info(f"{conn} with session id {client._SID}: retrieving period {period.start} - {period.end}")

        counter = 0
        saved_files = list()
        for entry in result:
            save_file = f"{save_file_prefix}-{counter}.xml"
            write_to_file(entry, save_file)
            counter += 1
            saved_files.append(save_file)

        return saved_files

    @staticmethod
    @backoff.on_exception(
        backoff.constant, WebFault, max_tries=WosUtilConst.RETRIES, interval=WosUtilConst.SESSION_CALL_PERIOD
    )
    def download_wos_batch(
        login: str, password: str, batch: list, conn: str, wos_inst_id: List[str], download_path: str
    ) -> List[str]:
        """Download one batch of WoS snapshots. Throttling limits are more conservative than WoS limits.
        Throttle limits may or may not be enforced. Probably depends on how executors spin up tasks.

        :param login: login.
        :param password: password.
        :param batch: List of tuples of (start_date, end_date) to fetch.
        :param conn: connection_id string from Airflow variable.
        :param wos_inst_id: List of Institutional ID to query, e.g, "Curtin University"
        :param download_path: download path to save response to.
        :return: List of saved files from this batch.
        """

        file_list = list()
        with WosClient(login, password) as client:
            for period in batch:
                files = WosUtility.download_wos_period(client, conn, period, wos_inst_id, download_path)
                file_list += files
        return file_list

    @staticmethod
    def download_wos_parallel(
        login: str, password: str, schedule: list, conn: str, wos_inst_id: List[str], download_path: str
    ) -> List[str]:
        """Download WoS snapshot with parallel sessions. Using threads.

        :param login: WoS login
        :param password: WoS password
        :param schedule: List of date range (start_date, end_date) tuples to download.
        :param conn: Airflow connection_id string.
        :param wos_inst_id: List of Institutional ID to query, e.g, "Curtin University"
        :param download_path: Path to download to.
        :return: List of files downloaded.
        """

        sessions = WosUtilConst.SESSION_CALL_LIMIT
        batch_size = ceil(len(schedule) / sessions)

        # Evenly distribute schedule among the session batches. Last session gets whatever the remaining tail is.
        batches = [schedule[i * batch_size : (i + 1) * batch_size] for i in range(sessions - 1)]
        batches.append(schedule[(sessions - 1) * batch_size :])  # Last session

        file_list = []
        with ThreadPoolExecutor(max_workers=sessions) as executor:
            futures = []
            for i in range(sessions):
                futures.append(
                    executor.submit(
                        WosUtility.download_wos_batch, login, password, batches[i], conn, wos_inst_id, download_path
                    )
                )
            for future in as_completed(futures):
                files = future.result()
                file_list = file_list + files

        return file_list

    @staticmethod
    def download_wos_sequential(
        login: str, password: str, schedule: list, conn: str, wos_inst_id: List[str], download_path: str
    ) -> List[str]:
        """Download WoS snapshot sequentially.

        :param login: WoS login
        :param password: WoS password
        :param schedule: List of date range (start_date, end_date) tuples to download.
        :param conn: Airflow connection_id string.
        :param wos_inst_id: List of Institutional ID to query, e.g, "Curtin University"
        :param download_path: Path to download to.
        :return: List of files downloaded.
        """

        return WosUtility.download_wos_batch(login, password, schedule, conn, wos_inst_id, download_path)

    @staticmethod
    def download_wos_snapshot(
        download_path: str, conn, wos_inst_id: List[str], end_date: pendulum.DateTime, mode: str
    ) -> List[str]:
        """Download snapshot from Web of Science for the given institution.

        :param download_path: the directory where the downloaded wos snapshot should be saved.
        :param conn: Airflow connection object.
        :param wos_inst_id: List of wos institution ids to use in query.
        :param end_date: end date of schedule. Usually the DAG start date of this DAG run.
        :param mode: Download mode to use. 'sequential' or 'parallel'
        """

        extra = json.loads(conn.extra)
        login = conn.login
        password = conn.password
        start_date = pendulum.parse(extra["start_date"])
        end_date = end_date
        schedule = build_schedule(start_date, end_date)

        if mode == "sequential" or len(schedule) <= WosUtilConst.SESSION_CALL_LIMIT:
            logging.info("Downloading snapshot with sequential method")
            return WosUtility.download_wos_sequential(login, password, schedule, str(conn), wos_inst_id, download_path)

        if mode == "parallel":
            logging.info("Downloading snapshot with parallel method")
            return WosUtility.download_wos_parallel(login, password, schedule, str(conn), wos_inst_id, download_path)

    @staticmethod
    def make_query(client: WosClient, query: OrderedDict) -> List[Any]:
        """Make the API calls to retrieve information from Web of Science.

        :param client: WosClient object.
        :param query: Constructed search query from use build_query.

        :return: List of XML responses.
        """

        results = WosUtility.wos_search(client, query)
        num_results = int(results.recordsFound)
        record_list = [results.records]

        if num_results > WosUtilConst.RESULT_LIMIT:
            for offset in range(2, num_results, WosUtilConst.RESULT_LIMIT):
                query["offset"] = offset
                record_list.append(WosUtility.wos_search(client, query).records)

        return record_list

    @staticmethod
    @sleep_and_retry
    @limits(calls=WosUtilConst.CALL_LIMIT, period=WosUtilConst.CALL_PERIOD)
    def wos_search(client: WosClient, query: OrderedDict) -> Any:
        """Throttling wrapper for the API call. This is a global limit for this API when called from a program on the
        same machine. If you are throttled, it will throw a WebFault and the exception message will contain the phrase
        'Request denied by Throttle server'

        Limiting to 1 call per second even though theoretical limit is 2 per second just in case.
        Throttle limits may or may not be enforced. Probably depends on how executors spin up tasks.

        :param client: WosClient object.
        :param query: Query object.
        :returns: Query results.
        """

        return client.search(**query)


class WosRelease:
    """Used to store info on a given WoS release.

    :param inst_id: institution id from the airflow connection (minus the wos_)
    :param wos_inst_id: List of institution ids to use in the WoS query.
    :param release_date: Release date (currently the execution date).
    :param dag_start: Start date of the dag (not execution date).
    :param project_id: The project id to use.
    :param download_bucket_name: Download bucket name to use for storing downloaded files in the cloud.
    :param transform_bucket_name: Transform bucket name to use for storing transformed files in the cloud.
    :param data_location: Location of the data servers
    """

    def __init__(
        self,
        inst_id: str,
        wos_inst_id: List[str],
        release_date: pendulum.DateTime,
        dag_start: pendulum.DateTime,
        project_id: str,
        download_bucket_name: str,
        transform_bucket_name: str,
        data_location: str,
        schema_ver: str,
    ):
        self.inst_id = inst_id
        self.wos_inst_id = wos_inst_id
        self.release_date = release_date
        self.dag_start = dag_start
        self.download_path = workflow_path(SubFolder.downloaded, WosTelescope.DAG_ID)
        self.transform_path = workflow_path(SubFolder.transformed, WosTelescope.DAG_ID)
        self.telescope_path = f"telescopes/{WosTelescope.DAG_ID}/{release_date}"
        self.project_id = project_id
        self.download_bucket_name = download_bucket_name
        self.transform_bucket_name = transform_bucket_name
        self.data_location = data_location
        self.schema_ver = schema_ver


class WosTelescope:
    """A container for holding the constants and static functions for the Web of Science telescope."""

    DAG_ID = "webofscience"
    ID_STRING_OFFSET = len(DAG_ID) + 1
    DESCRIPTION = "Web of Science: https://www.clarivate.com/webofsciencegroup"
    SCHEDULE_INTERVAL = "@monthly"
    QUEUE = "default"
    RETRIES = 3
    DATASET_ID = "clarivate"
    SCHEMA_PATH = "telescopes"
    TABLE_NAME = DAG_ID
    SCHEMA_VER = "wok5.4"

    TASK_ID_CHECK_DEPENDENCIES = "check_dependencies"
    TASK_CHECK_API_SERVER = "check_api_server"
    TASK_ID_DOWNLOAD = "download"
    TASK_ID_UPLOAD_DOWNLOADED = "upload_downloaded"
    TASK_ID_TRANSFORM_XML = "transform_xml"
    TASK_ID_TRANSFORM_DB_FORMAT = "transform_db_format"
    TASK_ID_UPLOAD_TRANSFORMED = "upload_transformed"
    TASK_ID_BQ_LOAD = "bq_load"
    TASK_ID_CLEANUP = "cleanup"

    XCOM_RELEASES = "releases"
    XCOM_DOWNLOAD_PATH = "download_path"
    XCOM_UPLOAD_ZIP_PATH = "download_zip_path"
    XCOM_JSON_PATH = "json_path"
    XCOM_JSON_HARVEST = "json_harvest"
    XCOM_JSONL_PATH = "jsonl_path"
    XCOM_JSONL_ZIP_PATH = "jsonl_zip_path"
    XCOM_JSONL_BLOB_PATH = "jsonl_blob_path"

    # Implement multiprocessing later if needed. WosClient seems to be synchronous i/o so asyncio doesn't help.
    DOWNLOAD_MODE = "sequential"  # Valid options: ['sequential', 'parallel']

    API_SERVER = "http://scientific.thomsonreuters.com"
    SCHEMA_ID_PREFIX = "http://scientific.thomsonreuters.com/schema/"

    @staticmethod
    def check_dependencies(**kwargs):
        """Check that all variables exist that are required to run the DAG.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        vars_valid = check_variables(
            AirflowVars.DATA_PATH,
            AirflowVars.PROJECT_ID,
            AirflowVars.DATA_LOCATION,
            AirflowVars.DOWNLOAD_BUCKET,
            AirflowVars.TRANSFORM_BUCKET,
        )
        if not vars_valid:
            raise AirflowException("Required variables are missing")

        conn = kwargs["conn"]

        # Validate extra field is set correctly.
        extra = conn.extra

        logging.info(f"Validating json in extra field of {conn}")
        try:
            extra_dict = json.loads(extra)
        except Exception as e:
            raise AirflowException(f"Error processing json extra fields in {conn} connection id profile: {e}")

        logging.info(f"Validating extra field keys for {conn}")

        # Check date is ok
        start_date = extra_dict["start_date"]
        if not validate_date(start_date):
            raise AirflowException(f"Invalid date string for {conn}: {start_date}")

        # Check institution id is set
        if "id" not in extra_dict:
            raise AirflowException(f'The "id" field is not set for {conn}.')

        # Check login is set
        if len(conn.login) == 0:
            raise AirflowException(f'The "login" field is not set for {conn}.')

        # Check password is set
        if len(conn.password) == 0:
            raise AirflowException(f'The "password" field is not set for {conn}.')

        logging.info(f"Checking for airflow override variables of {conn}")

        # Set project id override
        project_id = Variable.get(AirflowVars.PROJECT_ID)
        if "project_id" in extra_dict:
            project_id = extra_dict["project_id"]
            logging.info(f"Override for project_id found. Using: {project_id}")

        # Set download bucket name override
        download_bucket_name = Variable.get(AirflowVars.DOWNLOAD_BUCKET)
        if "download_bucket_name" in extra_dict:
            download_bucket_name = extra_dict["download_bucket_name"]
            logging.info(f"Override for download_bucket_name found. Using: {download_bucket_name}")

        # Set transform bucket name override
        transform_bucket_name = Variable.get(AirflowVars.TRANSFORM_BUCKET)
        if "transform_bucket_name" in extra_dict:
            transform_bucket_name = extra_dict["transform_bucket_name"]
            logging.info(f"Override for transform_bucket_name found. Using: {transform_bucket_name}")

        # Set data location override
        data_location = Variable.get(AirflowVars.DATA_LOCATION)
        if "data_location" in extra_dict:
            data_location = extra_dict["data_location"]
            logging.info(f"Override for data_location found. Using: {data_location}")

        # Push release information for other tasks
        wos_inst_id = get_as_list(extra_dict, "id")
        release = WosRelease(
            inst_id=kwargs["institution"],
            wos_inst_id=wos_inst_id,
            release_date=kwargs["execution_date"],
            dag_start=pendulum.parse(kwargs["dag_start"]),
            project_id=project_id,
            download_bucket_name=download_bucket_name,
            transform_bucket_name=transform_bucket_name,
            data_location=data_location,
            schema_ver=WosTelescope.SCHEMA_VER,
        )

        logging.info(
            f"WosRelease contains:\ndownload_bucket_name: {release.download_bucket_name}, transform_bucket_name: "
            + f"{release.transform_bucket_name}, data_location: {release.data_location}"
        )

        ti: TaskInstance = kwargs["ti"]
        ti.xcom_push(WosTelescope.XCOM_RELEASES, release)

    @staticmethod
    def check_api_server():
        """Check that http://scientific.thomsonreuters.com is still contactable.

        :return: the identifier of the task to execute next.
        """

        http_code_ok = 200

        logging.info(f"Checking API server {WosTelescope.API_SERVER} is up.")

        try:
            http_code = urllib.request.urlopen(WosTelescope.API_SERVER).getcode()
        except URLError as e:
            raise ValueError(f"Failed to fetch url because of: {e}")

        if http_code != http_code_ok:
            raise ValueError(f"HTTP response code {http_code} received.")

    @staticmethod
    def download(**kwargs):
        """Task to download the WoS snapshots.

        Pushes the following xcom:
            download_path (str): the path to a pickled file containing the list of xml responses in a month query.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        ti: TaskInstance = kwargs["ti"]
        release: WosRelease = WosTelescope.pull_release(ti)

        # Prepare paths
        download_files = WosUtility.download_wos_snapshot(
            release.download_path, kwargs["conn"], release.wos_inst_id, release.dag_start, WosTelescope.DOWNLOAD_MODE
        )

        # Notify next task of the files downloaded.
        ti.xcom_push(WosTelescope.XCOM_DOWNLOAD_PATH, download_files)

    @staticmethod
    def upload_downloaded(**kwargs):
        """Task to upload the downloaded WoS snapshots.

        Pushes the following xcom:
            upload_zip_path (str): the path to pickle zip file of downloaded response.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        ti: TaskInstance = kwargs["ti"]
        release: WosRelease = WosTelescope.pull_release(ti)

        # Pull messages
        download_list = ti.xcom_pull(
            key=WosTelescope.XCOM_DOWNLOAD_PATH, task_ids=WosTelescope.TASK_ID_DOWNLOAD, include_prior_dates=False
        )

        # Upload each snapshot
        logging.info("upload_downloaded: zipping and uploading downloaded files")
        zip_list = zip_files(download_list)
        upload_telescope_file_list(release.download_bucket_name, release.inst_id, release.telescope_path, zip_list)

        # Notify next task of the files downloaded.
        ti.xcom_push(WosTelescope.XCOM_UPLOAD_ZIP_PATH, zip_list)

    @staticmethod
    def transform_xml(**kwargs):
        """Task to transform the XML from the query to json.

        Pushes the following xcom:
            json_path (str): the path to json file of a converted xml response.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        ti: TaskInstance = kwargs["ti"]
        xml_files = ti.xcom_pull(
            key=WosTelescope.XCOM_DOWNLOAD_PATH, task_ids=WosTelescope.TASK_ID_DOWNLOAD, include_prior_dates=False
        )
        release: WosRelease = WosTelescope.pull_release(ti)

        # Process each xml file
        logging.info("transform_xml: transforming xml to dict and writing to json")
        json_file_list, schema_vers = write_xml_to_json(
            release.transform_path, release.release_date.isoformat(), release.inst_id, xml_files, WosUtility.parse_query
        )

        # Check we received consistent schema versions, and update release information.
        if schema_vers and schema_vers.count(WosTelescope.SCHEMA_VER) == len(schema_vers):
            release.schema_ver = schema_vers[0]
        else:
            raise AirflowException(f"Inconsistent schema versions received in response.")

        # Notify next task of the converted json files
        json_path = list()
        json_harvest = list()
        for file in json_file_list:
            file_name = os.path.basename(file)
            start = file_name.find("_") + 1
            end = file_name.rfind("-")
            harvest_datetime = file_name[start:end]
            json_path.append(file)
            json_harvest.append(harvest_datetime)
        ti.xcom_push(WosTelescope.XCOM_JSON_PATH, json_path)
        ti.xcom_push(WosTelescope.XCOM_JSON_HARVEST, json_harvest)

    @staticmethod
    def transform_db_format(**kwargs):
        """Task to transform the json into db field format (and in jsonlines form).

        Pushes the following xcom:
            version (str): the version of the GRID release.
            json_gz_file_name (str): the file name for the transformed GRID release.
            json_gz_file_path (str): the path to the transformed GRID release (including file name).

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        ti: TaskInstance = kwargs["ti"]
        release: WosRelease = WosTelescope.pull_release(ti)

        # Pull messages
        json_files = ti.xcom_pull(
            key=WosTelescope.XCOM_JSON_PATH, task_ids=WosTelescope.TASK_ID_TRANSFORM_XML, include_prior_dates=False
        )

        harvest_times = ti.xcom_pull(
            key=WosTelescope.XCOM_JSON_HARVEST, task_ids=WosTelescope.TASK_ID_TRANSFORM_XML, include_prior_dates=False
        )

        json_harvest_pair = list(zip(json_files, harvest_times))
        json_harvest_pair.sort()  # Sort by institution and date
        print(f"json harvest pairs are:\n{json_harvest_pair}")

        # Apply field extraction and transformation to jsonlines
        logging.info("transform_db_format: parsing and transforming into db format")
        jsonl_list = json_to_db(
            json_harvest_pair, release.release_date.isoformat(), WosJsonParser.parse_json, release.wos_inst_id
        )

        # Notify next task
        ti.xcom_push(WosTelescope.XCOM_JSONL_PATH, jsonl_list)

    @staticmethod
    def upload_transformed(**kwargs):
        """Task to upload the transformed WoS data into jsonlines files.

        Pushes the following xcom:
            release_date (str): the release date of the GRID release.
            blob_name (str): the name of the blob on the Google Cloud storage bucket.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        ti: TaskInstance = kwargs["ti"]
        release: WosRelease = WosTelescope.pull_release(ti)

        # Pull messages
        jsonl_paths = ti.xcom_pull(
            key=WosTelescope.XCOM_JSONL_PATH,
            task_ids=WosTelescope.TASK_ID_TRANSFORM_DB_FORMAT,
            include_prior_dates=False,
        )

        # Upload each snapshot
        logging.info("upload_transformed: zipping and uploading jsonlines to cloud")
        zip_list = zip_files(jsonl_paths)
        blob_list = upload_telescope_file_list(
            release.transform_bucket_name, release.inst_id, release.telescope_path, zip_list
        )

        # Notify next task of the files downloaded.
        ti.xcom_push(WosTelescope.XCOM_JSONL_BLOB_PATH, blob_list)
        ti.xcom_push(WosTelescope.XCOM_JSONL_ZIP_PATH, zip_list)

    @staticmethod
    def bq_load(**kwargs):
        """Task to load the transformed WoS snapshot into BigQuery.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        ti: TaskInstance = kwargs["ti"]
        release: WosRelease = WosTelescope.pull_release(ti)

        jsonl_zip_blobs = ti.xcom_pull(
            key=WosTelescope.XCOM_JSONL_BLOB_PATH,
            task_ids=WosTelescope.TASK_ID_UPLOAD_TRANSFORMED,
            include_prior_dates=False,
        )

        # Create dataset
        create_bigquery_dataset(
            release.project_id, WosTelescope.DATASET_ID, release.data_location, WosTelescope.DESCRIPTION
        )

        # Create table id
        table_id = bigquery_sharded_table_id(WosTelescope.TABLE_NAME, release.release_date)

        # Load into BigQuery
        analysis_schema_path = schema_folder()

        for file in jsonl_zip_blobs:
            schema_file_path = find_schema(
                analysis_schema_path, WosTelescope.TABLE_NAME, release.release_date, "", release.schema_ver
            )
            if schema_file_path is None:
                logging.error(
                    f"No schema found with search parameters: analysis_schema_path={WosTelescope.SCHEMA_PATH}, "
                    f"table_name={WosTelescope.TABLE_NAME}, release_date={release.release_date}, "
                    f"schema_ver={release.schema_ver}"
                )
                exit(os.EX_CONFIG)

            # Load BigQuery table
            uri = f"gs://{release.transform_bucket_name}/{file}"
            logging.info(f"URI: {uri}")

            load_bigquery_table(
                uri,
                WosTelescope.DATASET_ID,
                release.data_location,
                table_id,
                schema_file_path,
                SourceFormat.NEWLINE_DELIMITED_JSON,
                write_disposition=WriteDisposition.WRITE_APPEND,
            )

    @staticmethod
    def cleanup(**kwargs):
        """Delete files of downloaded, extracted and transformed releases.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        ti: TaskInstance = kwargs["ti"]

        delete_msg_files(ti, WosTelescope.XCOM_DOWNLOAD_PATH, WosTelescope.TASK_ID_DOWNLOAD)
        delete_msg_files(ti, WosTelescope.XCOM_UPLOAD_ZIP_PATH, WosTelescope.TASK_ID_UPLOAD_DOWNLOADED)
        delete_msg_files(ti, WosTelescope.XCOM_JSON_PATH, WosTelescope.TASK_ID_TRANSFORM_XML)
        delete_msg_files(ti, WosTelescope.XCOM_JSONL_PATH, WosTelescope.TASK_ID_TRANSFORM_DB_FORMAT)
        delete_msg_files(ti, WosTelescope.XCOM_JSONL_ZIP_PATH, WosTelescope.TASK_ID_UPLOAD_TRANSFORMED)

    @staticmethod
    def pull_release(ti: TaskInstance):
        """Get the WosRelease object from XCOM message."""
        return ti.xcom_pull(
            key=WosTelescope.XCOM_RELEASES, task_ids=WosTelescope.TASK_ID_CHECK_DEPENDENCIES, include_prior_dates=False
        )


class WosNameAttributes:
    """Helper class for parsing name attributes."""

    def __init__(self, data: dict):
        self._contribs = WosNameAttributes._get_contribs(data)

    @staticmethod
    def _get_contribs(data: dict) -> dict:
        """Helper function to parse the contributors structure to aid extraction of fields.

        :param data: dictionary to query.
        :return: Dictionary of attributes keyed by full_name string.
        """

        contrib_dict = dict()

        if "contributors" in data["static_data"]:
            contributors = get_as_list(data["static_data"]["contributors"], "contributor")
            for contributor in contributors:
                name_field = contributor["name"]
                first_name = name_field["first_name"]
                last_name = name_field["last_name"]
                attrib = dict()
                full_name = f"{first_name} {last_name}"
                if "@r_id" in name_field:
                    attrib["r_id"] = name_field["@r_id"]
                if "@orcid_id" in name_field:
                    attrib["orcid"] = name_field["@orcid_id"]
                contrib_dict[full_name] = attrib

        return contrib_dict

    def get_orcid(self, full_name: str) -> str:
        """Get the orcid id of a person. Note that full name must be the combination of first and last name.
        This is not necessarily the full_name field.

        :param full_name: The 'first_name last_name' string.
        :return: orcid id.
        """
        if full_name is None or full_name not in self._contribs or "orcid" not in self._contribs[full_name]:
            return None
        return self._contribs[full_name]["orcid"]

    def get_r_id(self, full_name: str) -> str:
        """Get the r_id of a person. Note that full name must be the combination of first and last name.
        This is not necessarily the full_name field.

        :param full_name: The 'first_name last_name' string.
        :return: r_id.
        """

        if full_name is None or full_name not in self._contribs or "r_id" not in self._contribs[full_name]:
            return None
        return self._contribs[full_name]["r_id"]


class WosJsonParser:
    """Helper methods to process the the converted json from Web of Science."""

    @staticmethod
    def get_identifiers(data: dict) -> Dict[str, Any]:
        """Extract identifier information.

        :param data: dictionary of web response.
        :return: Identifier record.
        """

        recognised_types = {
            "issn",
            "eissn",
            "isbn",
            "eisbn",
            "art_no",
            "meeting_abs",
            "xref_doi",
            "parent_book_doi",
            "doi",
        }
        field = dict()
        field["uid"] = data["UID"]

        for thing in recognised_types:
            field[thing] = None

        if "dynamic_data" not in data:
            return field

        if "cluster_related" not in data["dynamic_data"]:
            return field

        if "identifiers" not in data["dynamic_data"]["cluster_related"]:
            return field

        if data["dynamic_data"]["cluster_related"]["identifiers"] is None:
            return field

        if "identifier" not in data["dynamic_data"]["cluster_related"]["identifiers"]:
            return field

        identifiers = data["dynamic_data"]["cluster_related"]["identifiers"]
        identifier = get_as_list(identifiers, "identifier")

        for entry in identifier:
            type_ = entry["@type"]
            value = entry["@value"]

            if type_ in recognised_types:
                field[type_] = value

        return field

    @staticmethod
    def get_pub_info(data: dict) -> Dict[str, Any]:
        """Extract publication information fields.

        :param data: dictionary of web response.
        :return: Publication info record.
        """

        summary = data["static_data"]["summary"]

        field = dict()
        field["sort_date"] = None
        field["pub_type"] = None
        field["page_count"] = None
        field["source"] = None
        field["doc_type"] = None
        field["publisher"] = None
        field["publisher_city"] = None

        if "pub_info" in summary:
            pub_info = summary["pub_info"]
            field["sort_date"] = pub_info["@sortdate"]
            field["pub_type"] = pub_info["@pubtype"]
            field["page_count"] = int(pub_info["page"]["@page_count"])

        if "publishers" in summary and "publisher" in summary["publishers"]:
            publisher = summary["publishers"]["publisher"]
            field["publisher"] = publisher["names"]["name"]["full_name"]
            field["publisher_city"] = publisher["address_spec"]["city"]

        if "titles" in summary and "title" in summary["titles"]:
            titles = get_as_list(summary["titles"], "title")
            for title in titles:
                if title["@type"] == "source":
                    field["source"] = title["#text"]
                    break

        if "doctypes" in summary:
            doctypes = get_as_list(summary["doctypes"], "doctype")
            field["doc_type"] = doctypes[0]

        return field

    @staticmethod
    def get_title(data: dict) -> Union[None, str]:
        """Extract title. May raise exception on error.

        :param data: dictionary of web response.
        :return: String of title or None if not found.
        """
        if "titles" not in data["static_data"]["summary"]:
            return None

        for entry in data["static_data"]["summary"]["titles"]["title"]:
            if "@type" in entry and entry["@type"] == "item" and "#text" in entry:
                return entry["#text"]

        raise AirflowException("Schema change detected in title field. Please review.")

    @staticmethod
    def get_names(data: dict) -> List[Dict[str, Any]]:
        """Extract names fields.

        :param data: dictionary of web response.
        :return: List of name records.
        """

        field = list()
        if "names" not in data["static_data"]["summary"]:
            return field

        names = data["static_data"]["summary"]["names"]
        if "name" not in names:
            return field

        attrib = WosNameAttributes(data)
        names = get_as_list(data["static_data"]["summary"]["names"], "name")

        for name in names:
            entry = dict()
            entry["seq_no"] = int(get_entry_or_none(name, "@seq_no"))
            entry["role"] = get_entry_or_none(name, "@role")
            entry["first_name"] = get_entry_or_none(name, "first_name")
            entry["last_name"] = get_entry_or_none(name, "last_name")
            entry["wos_standard"] = get_entry_or_none(name, "wos_standard")
            entry["daisng_id"] = get_entry_or_none(name, "@daisng_id")
            entry["full_name"] = get_entry_or_none(name, "full_name")

            # Get around errors / booby traps for name retrieval
            first_name = entry["first_name"]
            last_name = entry["last_name"]
            full_name = f"{first_name} {last_name}"
            entry["orcid"] = attrib.get_orcid(full_name)
            entry["r_id"] = attrib.get_r_id(full_name)
            field.append(entry)

        return field

    @staticmethod
    def get_languages(data: dict) -> List[Dict[str, str]]:
        """Extract language fields.

        :param data: dictionary of web response.
        :return: List of language records.
        """

        lang_list = list()
        if "languages" not in data["static_data"]["fullrecord_metadata"]:
            return lang_list

        if "language" not in data["static_data"]["fullrecord_metadata"]["languages"]:
            return lang_list

        languages = get_as_list(data["static_data"]["fullrecord_metadata"]["languages"], "language")
        for entry in languages:
            lang_list.append({"type": entry["@type"], "name": entry["#text"]})
        return lang_list

    @staticmethod
    def get_refcount(data: dict) -> Union[int, None]:
        """Extract reference count.

        :param data: dictionary of web response.
        :return: Reference count.
        """

        if "refs" not in data["static_data"]["fullrecord_metadata"]:
            return None

        if "@count" not in data["static_data"]["fullrecord_metadata"]["refs"]:
            return None

        return int(data["static_data"]["fullrecord_metadata"]["refs"]["@count"])

    @staticmethod
    def get_abstract(data: dict) -> List[str]:
        """Extract abstracts.

        :param data: dictionary of web response.
        :return: List of abstracts.
        """

        abstract_list = list()
        if "abstracts" not in data["static_data"]["fullrecord_metadata"]:
            return abstract_list
        if "abstract" not in data["static_data"]["fullrecord_metadata"]["abstracts"]:
            return abstract_list

        abstracts = get_as_list(data["static_data"]["fullrecord_metadata"]["abstracts"], "abstract")
        for abstract in abstracts:
            texts = get_as_list(abstract["abstract_text"], "p")
            for text in texts:
                abstract_list.append(text)
        return abstract_list

    @staticmethod
    def get_keyword(data: dict) -> List[str]:
        """Extract keywords. Will also get the keywords from keyword plus if they are available.

        :param data: dictionary of web response.
        :return: List of keywords.
        """

        keywords = list()
        if "keywords" not in data["static_data"]["fullrecord_metadata"]:
            return keywords
        if "keyword" not in data["static_data"]["fullrecord_metadata"]["keywords"]:
            return keywords

        keywords = get_as_list(data["static_data"]["fullrecord_metadata"]["keywords"], "keyword")
        if "item" in data["static_data"] and "keywords_plus" in data["static_data"]["item"]:
            plus = get_as_list(data["static_data"]["item"]["keywords_plus"], "keyword")
            keywords = keywords + plus
        return keywords

    @staticmethod
    def get_conference(data: dict) -> List[Dict[str, Any]]:
        """Extract conference information.

        :param data: dictionary of web response.
        :return: List of conferences.
        """

        conferences = list()
        if "conferences" not in data["static_data"]["summary"]:
            return conferences
        if "conference" not in data["static_data"]["summary"]["conferences"]:
            return conferences

        conf_list = get_as_list(data["static_data"]["summary"]["conferences"], "conference")
        for conf in conf_list:
            conference = dict()
            conference["id"] = get_entry_or_none(conf, "@conf_id")
            if conference["id"] is not None:
                conference["id"] = int(conference["id"])
            conference["name"] = None

            if "conf_titles" in conf and "conf_title" in conf["conf_titles"]:
                titles = get_as_list(conf["conf_titles"], "conf_title")
                conference["name"] = titles[0]
            conferences.append(conference)
        return conferences

    @staticmethod
    def get_orgs(data: dict) -> list:
        """Extract the organisation information.

        :param data: dictionary of web response.
        :return: list of organisations or None
        """

        orgs = list()

        if "addresses" not in data["static_data"]["fullrecord_metadata"]:
            return orgs

        addr_list = get_as_list(data["static_data"]["fullrecord_metadata"]["addresses"], "address_name")
        if addr_list is None:
            return orgs

        for addr in addr_list:
            spec = addr["address_spec"]
            org = dict()

            org["city"] = get_entry_or_none(spec, "city")
            org["state"] = get_entry_or_none(spec, "state")
            org["country"] = get_entry_or_none(spec, "country")

            if "organizations" not in addr["address_spec"]:
                orgs.append(org)
                return orgs

            org_list = get_as_list(addr["address_spec"]["organizations"], "organization")
            for entry in org_list:
                if "@pref" in entry and entry["@pref"] == "Y":
                    org["org_name"] = entry["#text"]
                    break
            else:
                org["org_name"] = org_list[0]
                if not isinstance(org["org_name"], str):
                    raise AirflowException("Schema parsing error for org.")

            if "suborganizations" in addr["address_spec"]:
                org["suborgs"] = get_as_list(addr["address_spec"]["suborganizations"], "suborganization")

            if "names" in addr and "name" in addr["names"]:
                names = get_as_list(addr["names"], "name")
                names_list = list()
                for name in names:
                    entry = dict()
                    entry["first_name"] = get_entry_or_none(name, "first_name")
                    entry["last_name"] = get_entry_or_none(name, "last_name")
                    entry["daisng_id"] = get_entry_or_none(name, "@daisng_id")
                    entry["full_name"] = get_entry_or_none(name, "full_name")
                    entry["wos_standard"] = get_entry_or_none(name, "wos_standard")
                    names_list.append(entry)
                org["names"] = names_list
            orgs.append(org)
        return orgs

    @staticmethod
    def get_fund_ack(data: dict) -> dict:
        """Extract funding acknowledgements.

        :param data: dictionary of web response.
        :return: Funding acknowledgement information.
        """

        fund_ack = dict()
        fund_ack["text"] = list()
        fund_ack["grants"] = list()

        if "fund_ack" not in data["static_data"]["fullrecord_metadata"]:
            return fund_ack

        entry = data["static_data"]["fullrecord_metadata"]["fund_ack"]
        if "fund_text" in entry and "p" in entry["fund_text"]:
            fund_ack["text"] = get_as_list(entry["fund_text"], "p")

        if "grants" not in entry:
            return fund_ack

        grants = get_as_list(entry["grants"], "grant")
        for grant in grants:
            grant_info = dict()
            grant_info["agency"] = get_entry_or_none(grant, "grant_agency")
            grant_info["ids"] = list()
            if "grant_ids" in grant:
                grant_info["ids"] = get_as_list(grant["grant_ids"], "grant_id")
            fund_ack["grants"].append(grant_info)
        return fund_ack

    @staticmethod
    def get_categories(data: dict) -> dict:
        """Extract categories.

        :param data: dictionary of web response.
        :return: categories dictionary.
        """

        category_info = dict()
        if "category_info" not in data["static_data"]["fullrecord_metadata"]:
            return category_info

        entry = data["static_data"]["fullrecord_metadata"]["category_info"]
        category_info["headings"] = get_as_list_or_none(entry, "headings", "heading")
        category_info["subheadings"] = get_as_list_or_none(entry, "subheadings", "subheading")

        subject_list = list()
        subjects = get_as_list_or_none(entry, "subjects", "subject")
        for subject in subjects:
            subject_dict = dict()
            subject_dict["ascatype"] = get_entry_or_none(subject, "@ascatype")
            subject_dict["code"] = get_entry_or_none(subject, "@code")
            subject_dict["text"] = get_entry_or_none(subject, "#text")
            subject_list.append(subject_dict)
        category_info["subjects"] = subject_list

        return category_info

    @staticmethod
    def parse_json(data: dict, harvest_datetime: str, release_date: str, institutes: List[str]) -> dict:
        """Turn json data into db schema format.

        :param data: dictionary of web response.
        :param harvest_datetime: isoformat string of time the fetch took place.
        :param release_date: DAG execution date.
        :param institutes: List of institution ids used in the query.
        :return: dict of data in right field format.
        """

        entry = dict()
        entry["harvest_datetime"] = harvest_datetime
        entry["release_date"] = release_date
        entry["identifiers"] = WosJsonParser.get_identifiers(data)
        entry["pub_info"] = WosJsonParser.get_pub_info(data)
        entry["title"] = WosJsonParser.get_title(data)
        entry["names"] = WosJsonParser.get_names(data)
        entry["languages"] = WosJsonParser.get_languages(data)
        entry["ref_count"] = WosJsonParser.get_refcount(data)
        entry["abstract"] = WosJsonParser.get_abstract(data)
        entry["keywords"] = WosJsonParser.get_keyword(data)
        entry["conferences"] = WosJsonParser.get_conference(data)
        entry["fund_ack"] = WosJsonParser.get_fund_ack(data)
        entry["categories"] = WosJsonParser.get_categories(data)
        entry["orgs"] = WosJsonParser.get_orgs(data)
        entry["institution_ids"] = institutes

        return entry



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import glob
import os
import unittest
from typing import List
from zipfile import ZipFile

import natsort
import pendulum
from click.testing import CliRunner
from google.cloud import bigquery
from google.cloud import storage
from google.cloud.storage import Blob

from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.mag_telescope import (
    db_load_mag_release,
    list_mag_release_files,
    transform_mag_file,
    transform_mag_release,
)
from observatory.platform.utils.file_utils import _hash_file
from observatory.platform.utils.gc_utils import upload_files_to_cloud_storage
from observatory.platform.utils.test_utils import random_id


def extract_mag_release(file_path: str, unzip_path: str):
    """Extract a MAG release.

    :param file_path: the path to the archive to unzip.
    :param unzip_path: the path to unzip the files into. If the zip is of a folder, then the folder will be unzipped
    into this path.
    :return: None.
    """

    with ZipFile(file_path) as zip_file:
        zip_file.extractall(unzip_path)


class TestMagTelescope(unittest.TestCase):
    """ Tests for the functions used by the MAG telescope """

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestMagTelescope, self).__init__(*args, **kwargs)
        self.gc_project_id: str = os.getenv("TEST_GCP_PROJECT_ID")
        self.gc_bucket_name: str = os.getenv("TEST_GCP_BUCKET_NAME")
        self.gc_data_location: str = os.getenv("TEST_GCP_DATA_LOCATION")

        self.data_path = test_fixtures_folder("mag", "mag-2020-05-21.zip")
        self.release_date = pendulum.datetime(year=2020, month=5, day=21)
        self.release_folder = "mag-2020-05-21"
        self.extracted_folder = "extracted"
        self.transformed_folder = "transformed"
        self.release_folder = "mag-2020-05-21"
        self.extracted_folder = "extracted"
        self.transformed_folder = "transformed"
        self.folders = ["advanced", "mag", "nlp", "samples"]
        self.sub_folders = [
            ("Authors.txt_aes_tmp_2020-10-11_02-01-24", "Authors.txt"),
            ("PaperExtendedAttributes.txt_aes_tmp_2020-10-11_02-02-00", "PaperExtendedAttributes.txt"),
            ("PaperUrls.txt_aes_tmp_2020-10-11_02-01-43", "PaperUrls.txt"),
        ]
        self.advanced = [
            "EntityRelatedEntities.txt",
            "FieldOfStudyChildren.txt",
            "FieldOfStudyExtendedAttributes.txt",
            "FieldsOfStudy.txt",
            "PaperFieldsOfStudy.txt",
            "PaperRecommendations.txt",
            "RelatedFieldOfStudy.txt",
        ]
        self.mag = [
            "Affiliations.txt",
            "Authors.txt",
            "ConferenceInstances.txt",
            "ConferenceSeries.txt",
            "Journals.txt",
            "PaperAuthorAffiliations.txt",
            "PaperExtendedAttributes.txt",
            "PaperReferences.txt",
            "PaperUrls.txt",
            "Papers.txt",
        ]
        self.nlp = [
            "PaperAbstractsInvertedIndex.txt.1",
            "PaperAbstractsInvertedIndex.txt.2",
            "PaperCitationContexts.txt",
        ]
        self.samples = [
            "CreateDatabase.usql",
            "CreateFunctions.usql",
            "HIndexDatabricksSample.py",
            "ReadMe.pdf",
            "ReleaseNote.txt",
        ]

    def test_list_mag_release_files(self):
        """Test that list_mag_release_files lists all files in the MAG releases folder.
        :return: None.
        """

        with CliRunner().isolated_filesystem():
            # Make MAG folders
            folders = []
            for _, folder in enumerate(self.folders):
                path = os.path.join(self.release_folder, folder)
                os.makedirs(path, exist_ok=True)
                folders.append(path)

            # Make mag sub folders
            for sub_folder, sub_file in self.sub_folders:
                sub_folder_path = os.path.join(self.release_folder, "mag", sub_folder)
                os.makedirs(sub_folder_path, exist_ok=True)
                sub_file_path = os.path.join(sub_folder_path, sub_file)
                open(sub_file_path, "a").close()

            # advanced files
            expected_files = []
            for file_name in self.advanced:
                path = os.path.join(folders[0], file_name)
                open(path, "a").close()
                expected_files.append(path)

            # mag files
            for file_name in self.mag:
                path = os.path.join(folders[1], file_name)
                open(path, "a").close()
                expected_files.append(path)

            # nlp files
            for file_name in self.nlp:
                path = os.path.join(folders[2], file_name)
                open(path, "a").close()
                expected_files.append(path)

            # sample files
            for file_name in self.samples:
                path = os.path.join(folders[3], file_name)
                open(path, "a").close()

            # List MAG releases and check that output is as expected
            files = list_mag_release_files(self.release_folder)
            actual_files = [str(f) for f in files]
            self.assertListEqual(expected_files, actual_files)

        # Check that this function works with a folder of transformed files
        with CliRunner().isolated_filesystem():
            with CliRunner().isolated_filesystem():
                # Make MAG files
                file_names = self.advanced + self.mag + self.nlp
                expected_files = []
                os.makedirs(self.release_folder, exist_ok=True)
                for file_name in file_names:
                    path = os.path.join(self.release_folder, file_name)
                    open(path, "a").close()
                    expected_files.append(path)

                expected_files = sorted(expected_files)

                # List MAG releases and check that output is as expected
                files = list_mag_release_files(self.release_folder)
                actual_files = [str(f) for f in files]
                self.assertListEqual(expected_files, actual_files)

    def test_transform_mag_file(self):
        """Tests that transform_mag_file transforms a single file correctly.

        :return: None.
        """

        with CliRunner().isolated_filesystem():
            # Extract release zip file into folder
            extract_mag_release(self.data_path, self.extracted_folder)

            # Make input and output paths
            input_file_path = os.path.join(self.extracted_folder, self.release_folder, "mag", "Affiliations.txt")
            output_file_path = os.path.join(self.transformed_folder, self.release_folder, "mag")
            os.makedirs(output_file_path)
            output_file_path = os.path.join(output_file_path, "Affiliations.txt")

            # Transform file and check result
            result = transform_mag_file(input_file_path, output_file_path)
            self.assertTrue(result)
            expected_file_hash = "5570569e573a517587d3d11ec00eebf9"
            self.assertEqual(expected_file_hash, _hash_file(output_file_path, algorithm="md5"))

    def test_transform_mag_release(self):
        """Tests that transform_mag_release transforms an entire MAG release.

        :return: None.
        """

        with CliRunner().isolated_filesystem():
            # Make expected files
            expected_files = []
            for file in self.advanced + self.mag + self.nlp:
                expected_files.append(os.path.join(self.transformed_folder, self.release_folder, file))
            expected_files = natsort.natsorted(expected_files)

            # Extract release zip file into folder
            extract_mag_release(self.data_path, self.extracted_folder)

            # Transform release
            input_release_path = os.path.join(self.extracted_folder, self.release_folder)
            output_release_path = os.path.join(self.transformed_folder, self.release_folder)
            os.makedirs(output_release_path)
            result = transform_mag_release(input_release_path, output_release_path)

            # Test that the expected files exist
            self.assertTrue(result)
            actual_files = glob.glob(os.path.join(output_release_path, "**"))
            actual_files = natsort.natsorted(actual_files)
            self.assertEqual(expected_files, actual_files)

    def test_bq_load_mag_release(self):
        """Tests that db_load_mag_release successfully loads a MAG release into BigQuery.

        :return: None.
        """

        with CliRunner().isolated_filesystem():
            # Extract release zip file into folder
            extract_mag_release(self.data_path, self.extracted_folder)

            # Transform release
            input_release_path = os.path.join(self.extracted_folder, self.release_folder)
            output_release_path = os.path.join(self.transformed_folder, self.release_folder)
            os.makedirs(output_release_path)
            result = transform_mag_release(input_release_path, output_release_path)
            self.assertTrue(result)

            # Upload to cloud storage
            base_folder = random_id()
            print(f"base_folder: {base_folder}")
            release_path = f"{base_folder}/{self.release_folder}"
            posix_paths = list_mag_release_files(output_release_path)
            file_paths = [str(path) for path in posix_paths]
            blob_names = [f"{release_path}/{path.name}" for path in posix_paths]

            # Create random dataset id
            client = bigquery.Client()
            dataset_id = random_id()
            try:
                # Upload files to cloud storage
                result = upload_files_to_cloud_storage(self.gc_bucket_name, blob_names, file_paths)
                self.assertTrue(result)

                # Load release into BigQuery
                result = db_load_mag_release(
                    self.gc_project_id,
                    self.gc_bucket_name,
                    self.gc_data_location,
                    release_path,
                    self.release_date,
                    dataset_id=dataset_id,
                )

                # Check that all tables have loaded
                self.assertTrue(result)

                # Check that PaperAbstractsInvertedIndex has 100 rows, since it was loaded from two tables with 50
                # rows each
                table: bigquery.Table = client.get_table(f"{dataset_id}.PaperAbstractsInvertedIndex20200521")
                expected_num_rows = 100
                self.assertEqual(expected_num_rows, table.num_rows)
            finally:
                # Cleanup
                client.delete_dataset(dataset_id, delete_contents=True, not_found_ok=True)

                # Delete all blobs
                storage_client = storage.Client()
                bucket = storage_client.get_bucket(self.gc_bucket_name)
                blobs: List[Blob] = list(bucket.list_blobs(prefix=base_folder))
                for blob in blobs:
                    blob.delete()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import os
import shutil
from pathlib import Path

from recommonmark.transform import AutoStructify

from generate_schema_csv import generate_csv, generate_latest_files

# -- Project information -----------------------------------------------------

project = "Academic Observatory Workflows"
copyright = "2020-2021 Curtin University"
author = "Curtin University"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "pbr.sphinxext",
    "sphinx_rtd_theme",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "autoapi.extension",
    "recommonmark"
]

# Auto API settings: https://github.com/readthedocs/sphinx-autoapi
autoapi_type = "python"
autoapi_dirs = ["../academic_observatory_workflows"]
autoapi_add_toctree_entry = True
autoapi_python_use_implicit_namespaces = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_logo = "logo.png"
html_theme_options = {
    'logo_only': True,
    'display_version': False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["static"]


# recommonmark config, used to enable rst to be evaluated within markdown files
def setup(app):
    app.add_config_value("recommonmark_config", {"enable_eval_rst": True, "auto_toc_tree_section": "Contents"}, True)
    app.add_transform(AutoStructify)

generate_csv(schema_dir="../academic_observatory_workflows/database/schema")
generate_latest_files()
html_build_dir = "_build/html"
Path(html_build_dir).mkdir(exist_ok=True, parents=True)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import os

from observatory.platform.utils.config_utils import module_file_path


def test_fixtures_folder(*subdirs) -> str:
    """Get the path to the Academic Observatory Workflows test data directory.

    :return: the test data directory.
    """

    base_path = module_file_path("academic_observatory_workflows.fixtures")
    return os.path.join(base_path, *subdirs)


def schema_folder() -> str:
    """Return the path to the database schema template folder.

    :return: the path.
    """

    return module_file_path("academic_observatory_workflows.database.schema")


def sql_folder() -> str:
    """Return the path to the workflow SQL template folder.

    :return: the path.
    """

    return module_file_path("academic_observatory_workflows.database.sql")


def elastic_mappings_folder() -> str:
    """Get the Elasticsearch mappings path.

    :return: the elastic search schema path.
    """

    return module_file_path("academic_observatory_workflows.database.mappings")



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

"""
A DAG that harvests the Microsoft Academic Graph (MAG) dataset: https://www.microsoft.com/en-us/research/project/microsoft-academic-graph/

Saved to the BigQuery tables:
    <project_id>.mag.AffiliationsYYYYMMDD
    <project_id>.mag.AuthorExtendedAttributesYYYYMMDD
    <project_id>.mag.AuthorsYYYYMMDD
    <project_id>.mag.ConferenceInstancesYYYYMMDD
    <project_id>.mag.ConferenceSeriesYYYYMMDD
    <project_id>.mag.EntityRelatedEntitiesYYYYMMDD
    <project_id>.mag.FieldOfStudyChildrenYYYYMMDD
    <project_id>.mag.FieldOfStudyExtendedAttributesYYYYMMDD
    <project_id>.mag.FieldsOfStudyYYYYMMDD
    <project_id>.mag.JournalsYYYYMMDD
    <project_id>.mag.PaperAbstractsInvertedIndexYYYYMMDD
    <project_id>.mag.PaperAuthorAffiliationsYYYYMMDD
    <project_id>.mag.PaperCitationContextsYYYYMMDD
    <project_id>.mag.PaperExtendedAttributesYYYYMMDD
    <project_id>.mag.PaperFieldsOfStudyYYYYMMDD
    <project_id>.mag.PaperRecommendationsYYYYMMDD
    <project_id>.mag.PaperReferencesYYYYMMDD
    <project_id>.mag.PaperResourcesYYYYMMDD
    <project_id>.mag.PapersYYYYMMDD
    <project_id>.mag.PaperUrlsYYYYMMDD
    <project_id>.mag.RelatedFieldOfStudyYYYYMMDD
"""

import pendulum
from airflow import DAG
from airflow.operators.python_operator import PythonOperator, ShortCircuitOperator
from academic_observatory_workflows.workflows.mag_telescope import MagTelescope

default_args = {"owner": "airflow", "start_date": pendulum.datetime(2020, 7, 1)}

with DAG(dag_id=MagTelescope.DAG_ID, schedule_interval="@weekly", default_args=default_args, max_active_runs=1) as dag:
    # Check that dependencies exist before starting
    check = PythonOperator(
        task_id=MagTelescope.TASK_ID_CHECK_DEPENDENCIES,
        python_callable=MagTelescope.check_dependencies,
        provide_context=True,
        queue=MagTelescope.QUEUE,
    )

    # List releases and skip all subsequent tasks if there is no release to process
    list_releases = ShortCircuitOperator(
        task_id=MagTelescope.TASK_ID_LIST,
        python_callable=MagTelescope.list_releases,
        provide_context=True,
        queue=MagTelescope.QUEUE,
    )

    # Transfer all MAG releases to Google Cloud storage that were processed in the given interval
    transfer = PythonOperator(
        task_id=MagTelescope.TASK_ID_TRANSFER,
        python_callable=MagTelescope.transfer,
        provide_context=True,
        queue=MagTelescope.QUEUE,
    )

    # Download all MAG releases for a given interval
    download = PythonOperator(
        task_id=MagTelescope.TASK_ID_DOWNLOAD,
        python_callable=MagTelescope.download,
        provide_context=True,
        queue=MagTelescope.QUEUE,
    )

    # Transform all MAG releases for a given interval
    transform = PythonOperator(
        task_id=MagTelescope.TASK_ID_TRANSFORM,
        python_callable=MagTelescope.transform,
        provide_context=True,
        queue=MagTelescope.QUEUE,
    )

    # Upload all transformed MAG releases for a given interval to Google Cloud
    upload_transformed = PythonOperator(
        task_id=MagTelescope.TASK_ID_UPLOAD_TRANSFORMED,
        python_callable=MagTelescope.upload_transformed,
        provide_context=True,
        queue=MagTelescope.QUEUE,
        retries=MagTelescope.RETRIES,
    )

    # Load all MAG releases for a given interval to BigQuery
    bq_load = PythonOperator(
        task_id=MagTelescope.TASK_ID_BQ_LOAD,
        python_callable=MagTelescope.bq_load,
        provide_context=True,
        queue=MagTelescope.QUEUE,
    )

    # Cleanup local files
    cleanup = PythonOperator(
        task_id=MagTelescope.TASK_ID_CLEANUP,
        python_callable=MagTelescope.cleanup,
        provide_context=True,
        queue=MagTelescope.QUEUE,
    )

    check >> list_releases >> transfer >> download >> transform >> upload_transformed >> bq_load >> cleanup



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import unittest
from unittest.mock import patch

import pendulum
import vcr

from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.open_citations_telescope import (
    File,
    OpenCitationsRelease,
    fetch_open_citations_versions,
    list_open_citations_releases,
)


class TestOpenCitationsTelescope(unittest.TestCase):
    """Tests for the functions used by the Open Citations workflow"""

    def setUp(self) -> None:
        self.list_open_citations_releases_path = test_fixtures_folder(
            "open_citations", "list_open_citations_releases.yaml"
        )
        self.fetch_open_citations_versions_path = test_fixtures_folder(
            "open_citations", "fetch_open_citations_versions.yaml"
        )

    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_open_citations_release(self, mock_variable_get):
        """Test that the getter properties in OpenCitationsRelease work as expected.

        :return: None.
        """

        # Mock data variable
        data_path = "data"
        mock_variable_get.return_value = data_path

        release_date = pendulum.datetime(year=2018, month=7, day=5)
        file_name = "data.csv.zip"
        file = File(file_name, "", "")
        release = OpenCitationsRelease(release_date, [file])
        self.assertEqual(release.release_name, "open_citations_2018_07_05")
        self.assertEqual(release.download_path, "data/telescopes/download/open_citations/open_citations_2018_07_05")
        self.assertEqual(release.extract_path, "data/telescopes/extract/open_citations/open_citations_2018_07_05")
        self.assertEqual(release.transformed_blob_path, "telescopes/open_citations/open_citations_2018_07_05/*.csv")
        self.assertEqual(file.parent, release)
        self.assertEqual(file.download_blob_name, f"telescopes/open_citations/open_citations_2018_07_05/{file_name}")

    def test_list_open_citations_releases(self):
        """Test that the list_open_citations_releases function returns the correct results.

        :return: None.
        """

        with vcr.use_cassette(self.list_open_citations_releases_path):
            # No start date or end date
            expected_release_dates = [
                pendulum.datetime(year=2018, month=7, day=5),
                pendulum.datetime(year=2018, month=7, day=13),
                pendulum.datetime(year=2018, month=11, day=12),
                pendulum.datetime(year=2020, month=1, day=21),
                pendulum.datetime(year=2020, month=3, day=23),
                pendulum.datetime(year=2020, month=5, day=13),
                pendulum.datetime(year=2020, month=7, day=4),
                pendulum.datetime(year=2020, month=9, day=7),
            ]
            releases = list_open_citations_releases()
            self.assertEqual(len(releases), 8)
            for expected_date, release in zip(expected_release_dates, releases):
                self.assertEqual(expected_date.date(), release.release_date.date())

            # Start date
            expected_release_dates = [
                pendulum.datetime(year=2020, month=7, day=4),
                pendulum.datetime(year=2020, month=9, day=7),
            ]
            releases = list_open_citations_releases(start_date=pendulum.datetime(year=2020, month=7, day=1))
            self.assertEqual(len(releases), 2)
            for expected_date, release in zip(expected_release_dates, releases):
                self.assertEqual(expected_date.date(), release.release_date.date())

            # End date
            expected_release_dates = [
                pendulum.datetime(year=2018, month=7, day=5),
                pendulum.datetime(year=2018, month=7, day=13),
            ]
            releases = list_open_citations_releases(end_date=pendulum.datetime(year=2018, month=7, day=15))
            self.assertEqual(len(releases), 2)
            for expected_date, release in zip(expected_release_dates, releases):
                self.assertEqual(expected_date.date(), release.release_date.date())

            # Start date and end date
            expected_release_dates = [
                pendulum.datetime(year=2018, month=11, day=12),
                pendulum.datetime(year=2020, month=1, day=21),
                pendulum.datetime(year=2020, month=3, day=23),
            ]
            releases = list_open_citations_releases(
                start_date=pendulum.datetime(year=2018, month=11, day=1),
                end_date=pendulum.datetime(year=2020, month=4, day=1),
            )
            self.assertEqual(len(releases), 3)
            for expected_date, release in zip(expected_release_dates, releases):
                self.assertEqual(expected_date.date(), release.release_date.date())

    def test_fetch_open_citations_versions(self):
        """Test that the fetch_open_citations_versions function returns the correct results.

        :return: None.
        """

        with vcr.use_cassette(self.fetch_open_citations_versions_path):
            expected_versions = [
                {"url": "https://api.figshare.com/v2/articles/6741422/versions/1", "version": 1},
                {"url": "https://api.figshare.com/v2/articles/6741422/versions/2", "version": 2},
                {"url": "https://api.figshare.com/v2/articles/6741422/versions/3", "version": 3},
                {"url": "https://api.figshare.com/v2/articles/6741422/versions/4", "version": 4},
                {"url": "https://api.figshare.com/v2/articles/6741422/versions/5", "version": 5},
                {"url": "https://api.figshare.com/v2/articles/6741422/versions/6", "version": 6},
                {"url": "https://api.figshare.com/v2/articles/6741422/versions/7", "version": 7},
                {"url": "https://api.figshare.com/v2/articles/6741422/versions/8", "version": 8},
            ]
            versions = fetch_open_citations_versions()
            self.assertEqual(len(versions), 8)
            for expected_version, version in zip(expected_versions, versions):
                self.assertEqual(expected_version, version)



# Copyright 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Tuan Chien

from __future__ import annotations

import os
import random
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Tuple

import pandas as pd
import pendulum
from click.testing import CliRunner
from faker import Faker
from pendulum import DateTime

from academic_observatory_workflows.config import schema_folder, test_fixtures_folder
from observatory.platform.utils.file_utils import load_jsonl
from observatory.platform.utils.test_utils import Table, bq_load_tables

LICENSES = ["cc-by", None]

EVENT_TYPES = [
    "f1000",
    "stackexchange",
    "datacite",
    "twitter",
    "reddit-links",
    "wordpressdotcom",
    "plaudit",
    "cambia-lens",
    "hypothesis",
    "wikipedia",
    "reddit",
    "crossref",
    "newsfeed",
    "web",
]

OUTPUT_TYPES = [
    "journal_articles",
    "book_sections",
    "authored_books",
    "edited_volumes",
    "reports",
    "datasets",
    "proceedings_article",
    "other_outputs",
]

FUNDREF_COUNTRY_CODES = ["usa", "gbr", "aus", "can"]

FUNDREF_REGIONS = {"usa": "Americas", "gbr": "Europe", "aus": "Oceania", "can": "Americas"}

FUNDING_BODY_TYPES = [
    "For-profit companies (industry)",
    "Trusts, charities, foundations (both public and private)",
    "Associations and societies (private and public)",
    "National government",
    "Universities (academic only)",
    "International organizations",
    "Research institutes and centers",
    "Other non-profit organizations",
    "Local government",
    "Libraries and data archiving organizations",
]

FUNDING_BODY_SUBTYPES = {
    "For-profit companies (industry)": "pri",
    "Trusts, charities, foundations (both public and private)": "pri",
    "Associations and societies (private and public)": "pri",
    "National government": "gov",
    "Universities (academic only)": "gov",
    "International organizations": "pri",
    "Research institutes and centers": "pri",
    "Other non-profit organizations": "pri",
    "Local government": "gov",
    "Libraries and data archiving organizations": "gov",
}


@dataclass
class Institution:
    """An institution.

    :param id: unique identifier.
    :param name: the institution's name.
    :param grid_id: the institution's GRID id.
    :param country_code: the institution's country code.
    :param country_code_2: the institution's country code.
    :param subregion: the institution's subregion.
    :param papers: the papers published by the institution.
    :param types: the institution type.
    :param home_repo: the institution type.
    :param country: the institution country name.
    :param coordinates: the institution's coordinates.
    """

    id: int
    name: str = None
    grid_id: str = None
    country_code: str = None
    country_code_2: str = None
    region: str = None
    subregion: str = None
    papers: List[Paper] = None
    types: str = None
    home_repo: str = None
    country: str = None
    coordinates: str = None


def date_between_dates(start_ts: int, end_ts: int) -> DateTime:
    """Return a datetime between two timestamps.

    :param start_ts: the start timestamp.
    :param end_ts: the end timestamp.
    :return: the DateTime datetime.
    """

    r_ts = random.randint(start_ts, end_ts - 1)
    return pendulum.from_timestamp(r_ts)


@dataclass
class Paper:
    """A paper.

    :param id: unique identifier.
    :param doi: the DOI of the paper.
    :param title: the title of the paper.
    :param published_date: the date the paper was published.
    :param output_type: the output type, see OUTPUT_TYPES.
    :param authors: the authors of the paper.
    :param funders: the funders of the research published in the paper.
    :param journal: the journal this paper is published in.
    :param publisher: the publisher of this paper (the owner of the journal).
    :param events: a list of events related to this paper.
    :param cited_by: a list of papers that this paper is cited by.
    :param fields_of_study: a list of the fields of study of the paper.
    :param license:
    :param is_free_to_read_at_publisher:
    :param is_in_institutional_repo:
    """

    id: int
    doi: str = None
    title: str = None
    published_date: pendulum.Date = None
    output_type: str = None
    authors: List[Author] = None
    funders: List[Funder] = None
    journal: Journal = None
    publisher: Publisher = None
    events: List[Event] = None
    cited_by: List[Paper] = None
    fields_of_study: List[FieldOfStudy] = None
    license: str = None
    is_free_to_read_at_publisher: bool = False
    is_in_institutional_repo: bool = False

    @property
    def access_type(self) -> AccessType:
        """Return the access type for the paper.

        :return: AccessType.
        """

        gold_doaj = self.journal.license is not None
        gold = gold_doaj or (self.is_free_to_read_at_publisher and self.license is not None and not gold_doaj)
        hybrid = self.is_free_to_read_at_publisher and self.license is not None and not gold_doaj
        bronze = self.is_free_to_read_at_publisher and self.license is None and not gold_doaj
        green = self.is_in_institutional_repo
        green_only = self.is_in_institutional_repo and not gold_doaj and not self.is_free_to_read_at_publisher
        oa = gold or hybrid or bronze or green

        return AccessType(
            oa=oa, green=green, gold=gold, gold_doaj=gold_doaj, hybrid=hybrid, bronze=bronze, green_only=green_only
        )


@dataclass
class AccessType:
    """The access type of a paper.

    :param oa: whether the paper is open access or not.
    :param green: when the paper is available in an institutional repository.
    :param gold: when the paper is an open access journal or (it is not in an open access journal and is free to read
    at the publisher and has an open access license).
    :param gold_doaj: when the paper is an open access journal.
    :param hybrid: where the paper is free to read at the publisher, it has an open access license and the journal is
    not open access.
    :param bronze: when the paper is free to read at the publisher website however there is no license.
    :param green_only: where the paper is not free to read from the publisher, however it is available at an
    institutional repository.
    """

    oa: bool = None
    green: bool = None
    gold: bool = None
    gold_doaj: bool = None
    hybrid: bool = None
    bronze: bool = None
    green_only: bool = None


@dataclass
class Author:
    """An author.

    :param id: unique identifier.
    :param name: the name of the author.
    :param institution: the author's institution.
    """

    id: int
    name: str = None
    institution: Institution = None


@dataclass
class Funder:
    """A research funder.

    :param id: unique identifier.
    :param name: the name of the funder.
    :param doi: the DOI of the funder.
    :param country_code: the country code of the funder.
    :param region: the region the funder is located in.
    :param funding_body_type: the funding body type, see FUNDING_BODY_TYPES.
    :param funding_body_subtype: the funding body subtype, see FUNDING_BODY_SUBTYPES.
    """

    id: int
    name: str = None
    doi: str = None
    country_code: str = None
    region: str = None
    funding_body_type: str = None
    funding_body_subtype: str = None


@dataclass
class Publisher:
    """A publisher.

    :param id: unique identifier.
    :param name: the name of the publisher.
    :param doi_prefix: the publisher DOI prefix.
    :param journals: the journals owned by the publisher.
    """

    id: int
    name: str = None
    doi_prefix: int = None
    journals: List[Journal] = None


@dataclass
class FieldOfStudy:
    """A field of study.

    :param id: unique identifier.
    :param name: the field of study name.
    :param level: the field of study level.
    """

    id: int
    name: str = None
    level: int = None


@dataclass
class Journal:
    """A journal

    :param id: unique identifier.
    :param name: the journal name.
    :param name: the license that articles are published under by the journal.
    """

    id: int
    name: str = None
    license: str = None


@dataclass
class Event:
    """An event.

    :param source: the source of the event, see EVENT_TYPES.
    :param event_date: the date of the event.
    """

    source: str = None
    event_date: DateTime = None


InstitutionList = List[Institution]
AuthorList = List[Author]
FunderList = List[Funder]
PublisherList = List[Publisher]
PaperList = List[Paper]
FieldOfStudyList = List[FieldOfStudy]
EventsList = List[Event]


@dataclass
class ObservatoryDataset:
    """The generated observatory dataset.

    :param institutions: list of institutions.
    :param authors: list of authors.
    :param funders: list of funders.
    :param publishers: list of publishers.
    :param papers: list of papers.
    :param fields_of_study: list of fields of study.
    """

    institutions: InstitutionList
    authors: AuthorList
    funders: FunderList
    publishers: PublisherList
    papers: PaperList
    fields_of_study: FieldOfStudyList


def make_doi(doi_prefix: int):
    """Makes a randomised DOI given a DOI prefix.

    :param doi_prefix: the DOI prefix.
    :return: the DOI.
    """

    return f"10.{doi_prefix}/{str(uuid.uuid4())}"


def make_observatory_dataset(
    institutions: List[Institution],
    n_funders: int = 5,
    n_publishers: int = 5,
    n_authors: int = 10,
    n_papers: int = 50,
    n_fields_of_study_per_level: int = 5,
) -> ObservatoryDataset:
    """Generate an observatory dataset.

    :param institutions: a list of institutions.
    :param n_funders: the number of funders to generate.
    :param n_publishers: the number of publishers to generate.
    :param n_authors: the number of authors to generate.
    :param n_papers: the number of papers to generate.
    :param n_fields_of_study_per_level: the number of fields of study to generate per level.
    :return: the observatory dataset.
    """

    faker = Faker()
    funder_doi_prefix = 1000
    funders = make_funders(n_funders=n_funders, doi_prefix=funder_doi_prefix, faker=faker)
    publisher_doi_prefix = funder_doi_prefix + len(funders)
    publishers = make_publishers(n_publishers=n_publishers, doi_prefix=publisher_doi_prefix, faker=faker)
    fields_of_study = make_fields_of_study(n_fields_of_study_per_level=n_fields_of_study_per_level, faker=faker)
    authors = make_authors(n_authors=n_authors, institutions=institutions, faker=faker)
    papers = make_papers(
        n_papers=n_papers,
        authors=authors,
        funders=funders,
        publishers=publishers,
        fields_of_study=fields_of_study,
        faker=faker,
    )

    return ObservatoryDataset(institutions, authors, funders, publishers, papers, fields_of_study)


def make_funders(*, n_funders: int, doi_prefix: int, faker: Faker) -> FunderList:
    """Make the funders ground truth dataset.

    :param n_funders: number of funders to generate.
    :param doi_prefix: the DOI prefix for the funders.
    :param faker: the faker instance.
    :return: a list of funders.
    """

    funders = []

    for i, _ in enumerate(range(n_funders)):
        country_code = random.choice(FUNDREF_COUNTRY_CODES)
        funding_body_type = random.choice(FUNDING_BODY_TYPES)
        funders.append(
            Funder(
                i,
                name=faker.company(),
                doi=make_doi(doi_prefix),
                country_code=country_code,
                region=FUNDREF_REGIONS[country_code],
                funding_body_type=funding_body_type,
                funding_body_subtype=FUNDING_BODY_SUBTYPES[funding_body_type],
            )
        )
        doi_prefix += 1

    return funders


def make_publishers(
    *,
    n_publishers: int,
    doi_prefix: int,
    faker: Faker,
    min_journals_per_publisher: int = 1,
    max_journals_per_publisher: int = 3,
) -> PublisherList:
    """Make publishers ground truth dataset.

    :param n_publishers: number of publishers.
    :param doi_prefix: the publisher DOI prefix.
    :param faker: the faker instance.
    :param min_journals_per_publisher: the min number of journals to generate per publisher.
    :param max_journals_per_publisher: the max number of journals to generate per publisher.
    :return:
    """

    publishers = []
    journal_id = 0
    for i, _ in enumerate(range(n_publishers)):
        n_journals_ = random.randint(min_journals_per_publisher, max_journals_per_publisher)
        journals_ = []
        for _ in range(n_journals_):
            journals_.append(Journal(journal_id, name=faker.company(), license=random.choice(LICENSES)))
            journal_id += 1

        publishers.append(Publisher(i, name=faker.company(), doi_prefix=doi_prefix, journals=journals_))
        doi_prefix += 1

    return publishers


def make_fields_of_study(
    *,
    n_fields_of_study_per_level: int,
    faker: Faker,
    n_levels: int = 6,
    min_title_length: int = 1,
    max_title_length: int = 3,
) -> FieldOfStudyList:
    """Generate the fields of study for the ground truth dataset.

    :param n_fields_of_study_per_level: the number of fields of study per level.
    :param faker: the faker instance.
    :param n_levels: the number of levels.
    :param min_title_length: the minimum field of study title length (words).
    :param max_title_length: the maximum field of study title length (words).
    :return: a list of the fields of study.
    """

    fields_of_study = []
    fos_id_ = 0
    for level in range(n_levels):
        for _ in range(n_fields_of_study_per_level):
            n_words_ = random.randint(min_title_length, max_title_length)
            name_ = faker.sentence(nb_words=n_words_)
            fos_ = FieldOfStudy(fos_id_, name=name_, level=level)
            fields_of_study.append(fos_)
            fos_id_ += 1

    return fields_of_study


def make_authors(*, n_authors: int, institutions: InstitutionList, faker: Faker) -> AuthorList:
    """Generate the authors ground truth dataset.

    :param n_authors: the number of authors to generate.
    :param institutions: the institutions.
    :param faker: the faker instance.
    :return: a list of authors.
    """

    authors = []
    for i, _ in enumerate(range(n_authors)):
        author = Author(i, name=faker.name(), institution=random.choice(institutions))
        authors.append(author)

    return authors


def make_papers(
    *,
    n_papers: int,
    authors: AuthorList,
    funders: FunderList,
    publishers: PublisherList,
    fields_of_study: List,
    faker: Faker,
    min_title_length: int = 2,
    max_title_length: int = 10,
    min_authors: int = 1,
    max_authors: int = 10,
    min_funders: int = 0,
    max_funders: int = 3,
    min_events: int = 0,
    max_events: int = 100,
    min_fields_of_study: int = 1,
    max_fields_of_study: int = 20,
) -> PaperList:
    """Generate the list of ground truth papers.

    :param n_papers: the number of papers to generate.
    :param authors: the authors list.
    :param funders: the funders list.
    :param publishers: the publishers list.
    :param fields_of_study: the fields of study list.
    :param faker: the faker instance.
    :param min_title_length: the min paper title length.
    :param max_title_length: the max paper title length.
    :param min_authors: the min number of authors for each paper.
    :param max_authors: the max number of authors for each paper.
    :param min_funders: the min number of funders for each paper.
    :param max_funders: the max number of funders for each paper.
    :param min_events: the min number of events per paper.
    :param max_events: the max number of events per paper.
    :param min_fields_of_study: the min fields of study per paper.
    :param max_fields_of_study: the max fields of study per paper.
    :return: the list of papers.
    """

    papers = []

    for i, _ in enumerate(range(n_papers)):
        # Random title
        n_words_ = random.randint(min_title_length, max_title_length)
        title_ = faker.sentence(nb_words=n_words_)

        # Random date
        published_date_ = pendulum.from_format(faker.date(), "YYYY-MM-DD").date()
        published_date_ = pendulum.date(year=published_date_.year, month=published_date_.month, day=published_date_.day)

        # Output type
        output_type_ = random.choice(OUTPUT_TYPES)

        # Pick a random list of authors
        n_authors_ = random.randint(min_authors, max_authors)
        authors_ = random.sample(authors, n_authors_)

        # Random funder
        n_funders_ = random.randint(min_funders, max_funders)
        if n_funders_ > 0:
            funders_ = random.sample(funders, n_funders_)
        else:
            funders_ = []

        # Random publisher
        publisher_ = random.choice(publishers)

        # Journal
        journal_ = random.choice(publisher_.journals)

        # Random DOI
        doi_ = make_doi(publisher_.doi_prefix)

        # Random events
        n_events_ = random.randint(min_events, max_events)
        events_ = []
        today = datetime.now()
        today_ts = int(today.timestamp())
        start_date = datetime(today.year - 2, today.month, today.day)
        start_ts = int(start_date.timestamp())

        for _ in range(n_events_):
            event_date_ = date_between_dates(start_ts=start_ts, end_ts=today_ts)
            events_.append(Event(source=random.choice(EVENT_TYPES), event_date=event_date_))

        # Fields of study
        n_fos_ = random.randint(min_fields_of_study, max_fields_of_study)
        level_0_index = 199
        fields_of_study_ = [random.choice(fields_of_study[:level_0_index])]
        fields_of_study_.extend(random.sample(fields_of_study, n_fos_))

        # Open access status
        is_free_to_read_at_publisher_ = True
        if journal_.license is not None:
            # Gold
            license_ = journal_.license
        else:
            license_ = random.choice(LICENSES)
            if license_ is None:
                # Bronze: free to read on publisher website but no license
                is_free_to_read_at_publisher_ = bool(random.getrandbits(1))
            # Hybrid: license=True

        # Green: in a 'repository'
        is_in_institutional_repo_ = bool(random.getrandbits(1))
        # Green not bronze: Not free to read at publisher but in a 'repository'

        # Make paper
        paper = Paper(
            i,
            doi=doi_,
            title=title_,
            published_date=published_date_,
            output_type=output_type_,
            authors=authors_,
            funders=funders_,
            journal=journal_,
            publisher=publisher_,
            events=events_,
            fields_of_study=fields_of_study_,
            license=license_,
            is_free_to_read_at_publisher=is_free_to_read_at_publisher_,
            is_in_institutional_repo=is_in_institutional_repo_,
        )
        papers.append(paper)

    # Create paper citations
    # Sort from oldest to newest
    papers.sort(key=lambda p: p.published_date)

    for i, paper in enumerate(papers):
        # Create cited_by
        n_papers_forwards = len(papers) - i
        n_cited_by = random.randint(0, int(n_papers_forwards / 2))
        paper.cited_by = random.sample(papers[i + 1 :], n_cited_by)

    return papers


def make_open_citations(dataset: ObservatoryDataset) -> List[Dict]:
    """Generate an Open Citations table from an ObservatoryDataset instance.

    :param dataset: the Observatory Dataset.
    :return: table rows.
    """

    records = []

    def make_oc_timespan(cited_date: pendulum.Date, citing_date: pendulum.Date):
        ts = "P"
        delta = citing_date - cited_date
        years = delta.in_years()
        months = delta.in_months() - years * 12

        if years > 0:
            ts += f"{years}Y"

        if months > 0 or years == 0:
            ts += f"{months}M"

        return ts

    def is_author_sc(cited_: Paper, citing_: Paper):
        for cited_author in cited_.authors:
            for citing_author in citing_.authors:
                if cited_author.name == citing_author.name:
                    return True
        return False

    def is_journal_sc(cited_: Paper, citing_: Paper):
        return cited_.journal.name == citing_.journal.name

    for cited in dataset.papers:
        for citing in cited.cited_by:
            records.append(
                {
                    "oci": "",
                    "citing": citing.doi,
                    "cited": cited.doi,
                    "creation": citing.published_date.strftime("%Y-%m"),
                    "timespan": make_oc_timespan(cited.published_date, citing.published_date),
                    "journal_sc": is_author_sc(cited, citing),
                    "author_sc": is_journal_sc(cited, citing),
                }
            )

    return records


def make_crossref_events(dataset: ObservatoryDataset) -> List[Dict]:
    """Generate the Crossref Events table from an ObservatoryDataset instance.

    :param dataset: the Observatory Dataset.
    :return: table rows.
    """

    events = []

    for paper in dataset.papers:
        for event in paper.events:
            obj_id = f"https://doi.org/{paper.doi}"
            occurred_at = f"{event.event_date.to_datetime_string()} UTC"
            source_id = event.source
            events.append(
                {
                    "obj_id": obj_id,
                    "timestamp": occurred_at,
                    "occurred_at": occurred_at,
                    "source_id": source_id,
                    "id": str(uuid.uuid4()),
                }
            )

    return events


def make_unpaywall(dataset: ObservatoryDataset) -> List[Dict]:
    """Generate the Unpaywall table from an ObservatoryDataset instance.

    :param dataset: the Observatory Dataset.
    :return: table rows.
    """

    records = []
    genre_lookup = {
        "journal_articles": ["journal-article"],
        "book_sections": ["book-section", "book-part", "book-chapter"],
        "authored_books": ["book", "monograph"],
        "edited_volumes": ["edited-book"],
        "reports": ["report"],
        "datasets": ["dataset"],
        "proceedings_article": ["proceedings-article"],
        "other_outputs": ["other-outputs"],
    }

    for paper in dataset.papers:
        # Make OA status
        journal_is_in_doaj = paper.journal.license is not None

        oa_locations = []
        if paper.is_free_to_read_at_publisher:
            oa_location = {"host_type": "publisher", "license": paper.license, "url": ""}
            oa_locations.append(oa_location)

        if paper.is_in_institutional_repo:
            oa_location = {"host_type": "repository", "license": paper.license, "url": ""}
            oa_locations.append(oa_location)

        is_oa = len(oa_locations) > 0
        if is_oa:
            best_oa_location = oa_locations[0]
        else:
            best_oa_location = None

        # Create record
        records.append(
            {
                "doi": paper.doi,
                "year": paper.published_date.year,
                "genre": random.choice(genre_lookup[paper.output_type]),
                "publisher": paper.publisher.name,
                "journal_name": paper.journal.name,
                "is_oa": is_oa,
                "journal_is_in_doaj": journal_is_in_doaj,
                "best_oa_location": best_oa_location,
                "oa_locations": oa_locations,
            }
        )

    return records


@dataclass
class MagDataset:
    """A container to hold the Microsoft Academic Graph tables.

    :param: Affiliations table rows.
    :param: Papers table rows.
    :param: PaperAuthorAffiliations rows.
    :param: FieldsOfStudy rows.
    :param: PaperFieldsOfStudy rows.
    """

    affiliations: List[Dict]
    papers: List[Dict]
    paper_author_affiliations: List[Dict]
    fields_of_study: List[Dict]
    paper_fields_of_study: List[Dict]


def make_mag(dataset: ObservatoryDataset) -> MagDataset:
    """Generate the Microsoft Academic Graph tables from an ObservatoryDataset instance.

    :param dataset: the Observatory Dataset.
    :return: the Microsoft Academic Graph dataset.
    """

    # Create affiliations
    affiliations = []
    for institute in dataset.institutions:
        affiliations.append({"AffiliationId": institute.id, "DisplayName": institute.name, "GridId": institute.grid_id})

    # Create fields of study
    fields_of_study = []
    for fos in dataset.fields_of_study:
        fields_of_study.append({"FieldOfStudyId": fos.id, "DisplayName": fos.name, "Level": fos.level})

    # Create papers, paper_author_affiliations and paper_fields_of_study
    papers = []
    paper_author_affiliations = []
    paper_fields_of_study = []
    for paper in dataset.papers:
        papers.append({"PaperId": paper.id, "CitationCount": len(paper.cited_by), "Doi": paper.doi})

        for author in paper.authors:
            paper_author_affiliations.append(
                {"PaperId": paper.id, "AuthorId": author.id, "AffiliationId": author.institution.id}
            )

        for fos in paper.fields_of_study:
            paper_fields_of_study.append({"PaperId": paper.id, "FieldOfStudyId": fos.id})

    return MagDataset(affiliations, papers, paper_author_affiliations, fields_of_study, paper_fields_of_study)


def make_crossref_fundref(dataset: ObservatoryDataset) -> List[Dict]:
    """Generate the Crossref Fundref table from an ObservatoryDataset instance.

    :param dataset: the Observatory Dataset.
    :return: table rows.
    """

    records = []

    for funder in dataset.funders:
        records.append(
            {
                "pre_label": funder.name,
                "funder": f"http://dx.doi.org/{funder.doi}",
                "country_code": funder.country_code,
                "region": funder.region,
                "funding_body_type": funder.funding_body_type,
                "funding_body_sub_type": funder.funding_body_subtype,
            }
        )

    return records


def make_crossref_metadata(dataset: ObservatoryDataset) -> List[Dict]:
    """Generate the Crossref Metadata table from an ObservatoryDataset instance.

    :param dataset: the Observatory Dataset.
    :return: table rows.
    """

    records = []

    for paper in dataset.papers:
        # Create funders
        funders = []
        for funder in paper.funders:
            funders.append({"name": funder.name, "DOI": funder.doi, "award": None, "doi_asserted_by": None})

        # Add Crossref record
        records.append(
            {
                "title": [paper.title],
                "DOI": paper.doi,
                "is_referenced_by_count": len(paper.cited_by),
                "issued": {
                    "date_parts": [paper.published_date.year, paper.published_date.month, paper.published_date.day]
                },
                "funder": funders,
                "publisher": paper.publisher.name,
            }
        )

    return records


def bq_load_observatory_dataset(
    observatory_dataset: ObservatoryDataset,
    bucket_name: str,
    dataset_id_all: str,
    dataset_id_settings: str,
    release_date: DateTime,
    data_location: str,
):
    """Load the fake Observatory Dataset in BigQuery.

    :param observatory_dataset: the Observatory Dataset.
    :param bucket_name: the Google Cloud Storage bucket name.
    :param dataset_id_all: the dataset id for all data tables.
    :param dataset_id_settings: the dataset id for settings tables.
    :param release_date: the release date for the observatory dataset.
    :param data_location: the location of the BigQuery dataset.
    :return: None.
    """

    # Generate source datasets
    open_citations = make_open_citations(observatory_dataset)
    crossref_events = make_crossref_events(observatory_dataset)
    mag: MagDataset = make_mag(observatory_dataset)
    crossref_fundref = make_crossref_fundref(observatory_dataset)
    unpaywall = make_unpaywall(observatory_dataset)
    crossref_metadata = make_crossref_metadata(observatory_dataset)

    # Load fake GRID and settings datasets
    test_doi_path = test_fixtures_folder("doi")
    grid = load_jsonl(os.path.join(test_doi_path, "grid.jsonl"))
    iso3166_countries_and_regions = load_jsonl(os.path.join(test_doi_path, "iso3166_countries_and_regions.jsonl"))
    grid_to_home_url = load_jsonl(os.path.join(test_doi_path, "grid_to_home_url.jsonl"))
    groupings = load_jsonl(os.path.join(test_doi_path, "groupings.jsonl"))
    mag_affiliation_override = load_jsonl(os.path.join(test_doi_path, "mag_affiliation_override.jsonl"))

    analysis_schema_path = schema_folder()
    with CliRunner().isolated_filesystem() as t:
        tables = [
            Table("crossref_events", False, dataset_id_all, crossref_events, "crossref_events", analysis_schema_path),
            Table(
                "crossref_metadata", True, dataset_id_all, crossref_metadata, "crossref_metadata", analysis_schema_path
            ),
            Table("crossref_fundref", True, dataset_id_all, crossref_fundref, "crossref_fundref", analysis_schema_path),
            Table("Affiliations", True, dataset_id_all, mag.affiliations, "MagAffiliations", analysis_schema_path),
            Table("FieldsOfStudy", True, dataset_id_all, mag.fields_of_study, "MagFieldsOfStudy", analysis_schema_path),
            Table(
                "PaperAuthorAffiliations",
                True,
                dataset_id_all,
                mag.paper_author_affiliations,
                "MagPaperAuthorAffiliations",
                analysis_schema_path,
            ),
            Table(
                "PaperFieldsOfStudy",
                True,
                dataset_id_all,
                mag.paper_fields_of_study,
                "MagPaperFieldsOfStudy",
                analysis_schema_path,
            ),
            Table("Papers", True, dataset_id_all, mag.papers, "MagPapers", analysis_schema_path),
            Table("open_citations", True, dataset_id_all, open_citations, "open_citations", analysis_schema_path),
            Table("unpaywall", True, dataset_id_all, unpaywall, "unpaywall", analysis_schema_path),
            Table("grid", True, dataset_id_all, grid, "grid", analysis_schema_path),
            Table(
                "iso3166_countries_and_regions",
                False,
                dataset_id_all,
                iso3166_countries_and_regions,
                "iso3166_countries_and_regions",
                analysis_schema_path,
            ),
            Table(
                "grid_to_home_url",
                False,
                dataset_id_settings,
                grid_to_home_url,
                "grid_to_home_url",
                analysis_schema_path,
            ),
            Table("groupings", False, dataset_id_settings, groupings, "groupings", analysis_schema_path),
            Table(
                "mag_affiliation_override",
                False,
                dataset_id_settings,
                mag_affiliation_override,
                "mag_affiliation_override",
                analysis_schema_path,
            ),
            Table(
                "PaperAbstractsInvertedIndex",
                True,
                dataset_id_all,
                [],
                "MagPaperAbstractsInvertedIndex",
                analysis_schema_path,
            ),
            Table("Journals", True, dataset_id_all, [], "MagJournals", analysis_schema_path),
            Table("ConferenceInstances", True, dataset_id_all, [], "MagConferenceInstances", analysis_schema_path),
            Table("ConferenceSeries", True, dataset_id_all, [], "MagConferenceSeries", analysis_schema_path),
            Table(
                "FieldOfStudyExtendedAttributes",
                True,
                dataset_id_all,
                [],
                "MagFieldOfStudyExtendedAttributes",
                analysis_schema_path,
            ),
            Table(
                "PaperExtendedAttributes", True, dataset_id_all, [], "MagPaperExtendedAttributes", analysis_schema_path
            ),
            Table("PaperResources", True, dataset_id_all, [], "MagPaperResources", analysis_schema_path),
            Table("PaperUrls", True, dataset_id_all, [], "MagPaperUrls", analysis_schema_path),
            Table("PaperMeSH", True, dataset_id_all, [], "MagPaperMeSH", analysis_schema_path),
            Table("orcid", False, dataset_id_all, [], "orcid", analysis_schema_path),
        ]

        bq_load_tables(tables=tables, bucket_name=bucket_name, release_date=release_date, data_location=data_location)


def aggregate_events(events: List[Event]) -> Tuple[List[Dict], List[Dict], List[Dict]]:
    """Aggregate events by source into total events for all time, monthly and yearly counts.

    :param events: list of events.
    :return: list of events for each source aggregated by all time, months and years.
    """

    lookup_totals = dict()
    lookup_months = dict()
    lookup_years = dict()
    for event in events:
        # Total events
        if event.source in lookup_totals:
            lookup_totals[event.source] += 1
        else:
            lookup_totals[event.source] = 1

        # Events by month
        month = event.event_date.strftime("%Y-%m")
        month_key = (event.source, month)
        if month_key in lookup_months:
            lookup_months[month_key] += 1
        else:
            lookup_months[month_key] = 1

        # Events by year
        year = event.event_date.year
        year_key = (event.source, year)
        if year_key in lookup_years:
            lookup_years[year_key] += 1
        else:
            lookup_years[year_key] = 1

    total = [{"source": source, "count": count} for source, count in lookup_totals.items()]
    months = [{"source": source, "month": month, "count": count} for (source, month), count in lookup_months.items()]
    years = [{"source": source, "year": year, "count": count} for (source, year), count in lookup_years.items()]

    # Sort
    sort_events(total, months, years)
    return total, months, years


def sort_events(events: List[Dict], months: List[Dict], years: List[Dict]):
    """Sort events in-place.

    :param events: events all time.
    :param months: events by month.
    :param years: events by year.
    :return: None.
    """

    events.sort(key=lambda x: x["source"])
    months.sort(key=lambda x: f"{x['month']}{x['source']}{x['count']}")
    years.sort(key=lambda x: f"{x['year']}{x['source']}{x['count']}")


def make_doi_table(dataset: ObservatoryDataset) -> List[Dict]:
    """Generate the DOI table from an ObservatoryDataset instance.

    :param dataset: the Observatory Dataset.
    :return: table rows.
    """

    records = []
    for paper in dataset.papers:
        # Doi, events and grids
        doi = paper.doi.upper()
        events = make_doi_events(doi, paper.events)
        grids = make_doi_grids(paper.authors)

        # Affiliations: institutions, countries, regions, subregion, funders, journals, publishers
        institutions = make_doi_institutions(paper.authors)
        countries = make_doi_countries(paper.authors)
        regions = make_doi_regions(paper.authors)
        subregions = make_doi_subregions(paper.authors)
        funders = make_doi_funders(paper.funders)
        journals = make_doi_journals(paper.journal)
        publishers = make_doi_publishers(paper.publisher)

        # Make final record
        records.append(
            {
                "doi": doi,
                "crossref": {
                    "title": paper.title,
                    "published_year": paper.published_date.year,
                    "published_month": paper.published_date.month,
                    "published_year_month": f"{paper.published_date.year}-{paper.published_date.month}",
                    "funder": [{"name": funder.name, "DOI": funder.doi} for funder in paper.funders],
                },
                "unpaywall": {},
                "unpaywall_history": {},
                "mag": {},
                "open_citations": {},
                "events": events,
                "grids": grids,
                "affiliations": {
                    "doi": doi,
                    "institutions": institutions,
                    "countries": countries,
                    "subregions": subregions,
                    "regions": regions,
                    "groupings": [],
                    "funders": funders,
                    "authors": [],
                    "journals": journals,
                    "publishers": publishers,
                },
            }
        )

    # Sort to match with sorted results
    records.sort(key=lambda r: r["doi"])

    return records


def make_doi_events(doi: str, event_list: EventsList) -> Dict:
    """Make the events for a DOI table row.

    :param doi: the DOI.
    :param event_list: a list of events for the paper.
    :return: the events for the DOI table.
    """

    events_total, events_months, events_years = aggregate_events(event_list)

    # When no events, events is None
    events = None
    if len(events_total):
        events = {
            "doi": doi,
            "events": events_total,
            "months": events_months,
            "years": events_years,
        }

    return events


def make_doi_grids(author_list: AuthorList) -> List[str]:
    """Make the grid ids for a DOI table row.

    :param author_list: the list of authors for the paper.
    :return: the grid ids for the paper.
    """

    return list(set([author.institution.grid_id for author in author_list]))


def make_doi_funders(funder_list: FunderList) -> List[Dict]:
    """Make a DOI table row funders affiliation list.

    :param funder_list: the funders list.
    :return: the funders affiliation list.
    """

    # Funders
    funders = {}
    for funder in funder_list:
        funders[funder.doi] = {
            "identifier": funder.name,
            "name": funder.name,
            "doi": funder.doi,
            "types": ["Funder"],
            "home_repo": [],
            "country": None,
            "country_code": funder.country_code,
            "country_code_2": None,
            "region": funder.region,
            "subregion": None,
            "coordinates": None,
            "funding_body_type": funder.funding_body_type,
            "funding_body_subtype": funder.funding_body_subtype,
            "members": [],
        }
    funders = [v for k, v in funders.items()]
    funders.sort(key=lambda x: x["identifier"])

    return funders


def make_doi_journals(journal: Journal) -> List[Dict]:
    """Make the journal affiliation list for a DOI table row.

    :param journal: the paper's journal.
    :return: the journal affiliation list.
    """

    return [
        {
            "identifier": journal.name,
            "types": ["Journal"],
            "name": journal.name,
            "home_repo": [],
            "country": None,
            "country_code": None,
            "country_code_2": None,
            "region": None,
            "subregion": None,
            "coordinates": None,
            "members": [],
        }
    ]


def to_affiliations_list(dict_: Dict):
    """Convert affiliation dict into a list.

    :param dict_: affiliation dict.
    :return: affiliation list.
    """

    l_ = []
    for k, v in dict_.items():
        v["members"] = list(v["members"])
        v["home_repo"] = list(v["home_repo"])
        v["members"].sort()
        if "count" in v:
            v["count"] = len(v["grids"])
            v.pop("grids", None)
        v["home_repo"].sort()
        l_.append(v)
    l_.sort(key=lambda x: x["identifier"])
    return l_


def make_doi_publishers(publisher: Publisher) -> List[Dict]:
    """Make the publisher affiliations for a DOI table row.

    :param publisher: the paper's publisher.
    :return: the publisher affiliations list.
    """

    return [
        {
            "identifier": publisher.name,
            "types": ["Publisher"],
            "name": publisher.name,
            "home_repo": [],
            "country": None,
            "country_code": None,
            "country_code_2": None,
            "region": None,
            "subregion": None,
            "coordinates": None,
            "members": [],
        }
    ]


def make_doi_institutions(author_list: AuthorList) -> List[Dict]:
    """Make the institution affiliations for a DOI table row.

    :param author_list: the paper's author list.
    :return: the institution affiliation list.
    """

    institutions = {}

    for author in author_list:
        # Institution
        inst = author.institution
        if inst.grid_id not in institutions:
            institutions[inst.grid_id] = {
                "identifier": inst.grid_id,
                "types": [inst.types],
                "name": inst.name,
                "home_repo": {inst.home_repo},
                "country": inst.country,
                "country_code": inst.country_code,
                "country_code_2": inst.country_code_2,
                "region": inst.region,
                "subregion": inst.subregion,
                "coordinates": inst.coordinates,
                "members": [],
            }

    return to_affiliations_list(institutions)


def make_doi_countries(author_list: AuthorList):
    """Make the countries affiliations for a DOI table row.

    :param author_list: the paper's author list.
    :return: the countries affiliation list.
    """

    countries = {}

    for author in author_list:
        inst = author.institution
        if inst.country not in countries:
            countries[inst.country] = {
                "identifier": inst.country_code,
                "name": inst.country,
                "types": ["Country"],
                "home_repo": {inst.home_repo},
                "country": inst.country,
                "country_code": inst.country_code,
                "country_code_2": inst.country_code_2,
                "region": inst.region,
                "subregion": inst.subregion,
                "coordinates": None,
                "count": 0,
                "members": {inst.grid_id},
                "grids": {inst.grid_id},
            }
        else:
            countries[inst.country]["members"].add(inst.grid_id)
            countries[inst.country]["home_repo"].add(inst.home_repo)
            countries[inst.country]["grids"].add(inst.grid_id)

    return to_affiliations_list(countries)


def make_doi_regions(author_list: AuthorList):
    """Make the regions affiliations for a DOI table row.

    :param author_list: the paper's author list.
    :return: the regions affiliation list.
    """

    regions = {}
    for author in author_list:
        inst = author.institution
        if inst.region not in regions:
            regions[inst.region] = {
                "identifier": inst.region,
                "name": inst.region,
                "types": ["Region"],
                "home_repo": {inst.home_repo},
                "country": None,
                "country_code": None,
                "country_code_2": None,
                "region": inst.region,
                "subregion": None,
                "coordinates": None,
                "count": 0,
                "members": {inst.subregion},
                "grids": {inst.grid_id},
            }
        else:
            regions[inst.region]["members"].add(inst.subregion)
            regions[inst.region]["home_repo"].add(inst.home_repo)
            regions[inst.region]["grids"].add(inst.grid_id)

    return to_affiliations_list(regions)


def make_doi_subregions(author_list: AuthorList):
    """Make the subregions affiliations for a DOI table row.

    :param author_list: the paper's author list.
    :return: the subregions affiliation list.
    """

    subregions = {}
    for author in author_list:
        inst = author.institution

        if inst.subregion not in subregions:
            subregions[inst.subregion] = {
                "identifier": inst.subregion,
                "name": inst.subregion,
                "types": ["Subregion"],
                "home_repo": {inst.home_repo},
                "country": None,
                "country_code": None,
                "country_code_2": None,
                "region": inst.region,
                "subregion": None,
                "coordinates": None,
                "count": 0,
                "members": {inst.country_code},
                "grids": {inst.grid_id},
            }
        else:
            subregions[inst.subregion]["members"].add(inst.country_code)
            subregions[inst.subregion]["home_repo"].add(inst.home_repo)
            subregions[inst.subregion]["grids"].add(inst.grid_id)

    return to_affiliations_list(subregions)


def calc_percent(value: float, total: float) -> float:
    """Calculate a percentage and round to 2dp.

    :param value: the value.
    :param total: the total.
    :return: the percentage.
    """

    return round(value / total * 100, 2)


def make_country_table(dataset: ObservatoryDataset) -> List[Dict]:
    """Generate the Observatory Country table from an ObservatoryDataset instance.

    :param dataset: the Observatory Dataset.
    :return: table rows.
    """

    data = []
    for paper in dataset.papers:
        for author in paper.authors:
            inst = author.institution
            at = paper.access_type
            data.append(
                {
                    "doi": paper.doi,
                    "id": inst.country_code,
                    "time_period": paper.published_date.year,
                    "name": inst.country,
                    "home_repo": None,
                    "country": inst.country,
                    "country_code": inst.country_code,
                    "country_code_2": inst.country_code_2,
                    "region": inst.region,
                    "subregion": inst.subregion,
                    "coordinates": None,
                    "total_outputs": 1,
                    "oa": at.oa,
                    "green": at.green,
                    "gold": at.gold,
                    "gold_doaj": at.gold_doaj,
                    "hybrid": at.hybrid,
                    "bronze": at.bronze,
                    "green_only": at.green_only,
                }
            )

    df = pd.DataFrame(data)
    df.drop_duplicates(inplace=True)
    agg = {
        "id": "first",
        "time_period": "first",
        "name": "first",
        "home_repo": "first",
        "country": "first",
        "country_code": "first",
        "country_code_2": "first",
        "region": "first",
        "subregion": "first",
        "coordinates": "first",
        "total_outputs": "sum",
        "oa": "sum",
        "green": "sum",
        "gold": "sum",
        "gold_doaj": "sum",
        "hybrid": "sum",
        "bronze": "sum",
        "green_only": "sum",
    }
    df = df.groupby(["id", "time_period"], as_index=False).agg(agg).sort_values(by=["id", "time_period"])
    records = []
    for i, row in df.iterrows():
        total_outputs = row["total_outputs"]
        oa = row["oa"]
        green = row["green"]
        gold = row["gold"]
        gold_doaj = row["gold_doaj"]
        hybrid = row["hybrid"]
        bronze = row["bronze"]
        green_only = row["green_only"]

        records.append(
            {
                "id": row["id"],
                "time_period": row["time_period"],
                "name": row["name"],
                "home_repo": row["home_repo"],
                "country": row["country"],
                "country_code": row["country_code"],
                "country_code_2": row["country_code_2"],
                "region": row["region"],
                "subregion": row["subregion"],
                "coordinates": row["coordinates"],
                "total_outputs": total_outputs,
                "access_types": {
                    "oa": {"total_outputs": oa, "percent": calc_percent(oa, total_outputs)},
                    "green": {"total_outputs": green, "percent": calc_percent(green, total_outputs)},
                    "gold": {"total_outputs": gold, "percent": calc_percent(gold, total_outputs)},
                    "gold_doaj": {"total_outputs": gold_doaj, "percent": calc_percent(gold_doaj, total_outputs)},
                    "hybrid": {"total_outputs": hybrid, "percent": calc_percent(hybrid, total_outputs)},
                    "bronze": {"total_outputs": bronze, "percent": calc_percent(bronze, total_outputs)},
                    "green_only": {"total_outputs": green_only, "percent": calc_percent(green_only, total_outputs)},
                },
                "citations": {},
                "output_types": [],
                "disciplines": {},
                "funders": [],
                "members": [],
                "publishers": [],
                "journals": [],
                "events": [],
            }
        )

    return records



# Copyright 2020-2021 Curtin University. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from observatory.platform.utils.config_utils import module_file_path
from academic_observatory_workflows.config import elastic_mappings_folder


class AcademicObservatoryWorkflowsConfig(unittest.TestCase):
    def test_elastic_schema_path(self):
        """ Test that the Elasticsearch schema path is correct """

        expected_path = module_file_path("academic_observatory_workflows.database.mappings")
        actual_path = elastic_mappings_folder()
        self.assertEqual(expected_path, actual_path)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import logging
import os
import shutil
import unittest
from typing import List
from unittest.mock import patch

import pendulum
import vcr
from click.testing import CliRunner

from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.unpaywall_telescope import (
    UnpaywallRelease,
    UnpaywallTelescope,
    extract_release,
    list_releases,
    transform_release,
)
from observatory.platform.utils.file_utils import _hash_file
from observatory.platform.utils.workflow_utils import SubFolder, workflow_path


class TestUnpaywallTelescope(unittest.TestCase):
    """ Tests for the functions used by the unpaywall telescope """

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestUnpaywallTelescope, self).__init__(*args, **kwargs)

        # Unpaywall releases list
        self.list_unpaywall_releases_path = test_fixtures_folder("unpaywall", "list_unpaywall_releases.yaml")
        self.list_unpaywall_releases_hash = "78d1a129cb0aba072ca49e2599f60c10"

        # Unpaywall test release
        self.unpaywall_test_path = test_fixtures_folder("unpaywall", "unpaywall.jsonl.gz")
        self.unpaywall_test_file = "unpaywall_snapshot_3000-01-27T153236.jsonl.gz"
        self.unpaywall_test_url = (
            "https://unpaywall-data-snapshots.s3-us-west-2.amazonaws.com"
            "/unpaywall_snapshot_3000-01-27T153236.jsonl.gz"
        )
        self.unpaywall_test_date = pendulum.datetime(year=3000, month=1, day=27)
        self.unpaywall_test_download_file_name = "unpaywall_3000_01_27.jsonl.gz"
        self.unpaywall_test_decompress_file_name = "unpaywall_3000_01_27.jsonl"
        self.unpaywall_test_transform_file_name = "unpaywall_3000_01_27.jsonl"
        self.unpaywall_test_download_hash = "90051478f7b6689838d58edfc2450cb3"
        self.unpaywall_test_decompress_hash = "fe4e72ce54c4bb236802ddbb3dbee905"
        self.unpaywall_test_transform_hash = "62cbb5af5a78d2e0769a28d976971cba"
        self.start_date = pendulum.datetime(year=2018, month=3, day=29)
        self.end_date = pendulum.datetime(year=2020, month=4, day=29)

        # Turn logging to warning because vcr prints too much at info level
        logging.basicConfig()
        logging.getLogger().setLevel(logging.WARNING)

    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_list_releases(self, mock_variable_get):
        """Test that list releases returns a list of string with urls.

        :return: None.
        """

        data_path = "data"
        mock_variable_get.return_value = data_path

        with CliRunner().isolated_filesystem():
            with vcr.use_cassette(self.list_unpaywall_releases_path):
                releases = list_releases(self.start_date, self.end_date)
                self.assertIsInstance(releases, List)
                for release in releases:
                    self.assertIsInstance(release, UnpaywallRelease)
                self.assertEqual(13, len(releases))

    def test_parse_release_date(self):
        """Test that date obtained from url is string and in correct format.

        :return: None.
        """

        release_date = UnpaywallRelease.parse_release_date(self.unpaywall_test_file)
        self.assertEqual(self.unpaywall_test_date, release_date)

    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_filepath_download(self, mock_variable_get):
        """Test that path of downloaded file is correct for given url.

        :param home_mock: Mock observatory home path
        :return: None.
        """

        # Create data path and mock getting data path
        data_path = "data"
        mock_variable_get.return_value = data_path

        with CliRunner().isolated_filesystem():
            release = UnpaywallRelease(self.unpaywall_test_file, self.unpaywall_test_date, self.unpaywall_test_date)
            path = workflow_path(SubFolder.downloaded, UnpaywallTelescope.DAG_ID)
            self.assertEqual(os.path.join(path, self.unpaywall_test_download_file_name), release.filepath_download)

    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_filepath_extract(self, mock_variable_get):
        """Test that path of decompressed/extracted file is correct for given url.

        :param home_mock: Mock observatory home path
        :return: None.
        """

        # Create data path and mock getting data path
        data_path = "data"
        mock_variable_get.return_value = data_path

        with CliRunner().isolated_filesystem():
            release = UnpaywallRelease(self.unpaywall_test_file, self.unpaywall_test_date, self.unpaywall_test_date)
            path = workflow_path(SubFolder.extracted, UnpaywallTelescope.DAG_ID)
            self.assertEqual(os.path.join(path, self.unpaywall_test_decompress_file_name), release.filepath_extract)

    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_filepath_transform(self, mock_variable_get):
        """Test that path of transformed file is correct for given url.

        :param home_mock: Mock observatory home path
        :return: None.
        """

        # Create data path and mock getting data path
        data_path = "data"
        mock_variable_get.return_value = data_path

        with CliRunner().isolated_filesystem():
            release = UnpaywallRelease(self.unpaywall_test_file, self.unpaywall_test_date, self.unpaywall_test_date)
            path = workflow_path(SubFolder.transformed, UnpaywallTelescope.DAG_ID)
            self.assertEqual(os.path.join(path, self.unpaywall_test_transform_file_name), release.filepath_transform)

    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_extract_release(self, mock_variable_get):
        """Test that the release is decompressed as expected.

        :return: None.
        """

        # Create data path and mock getting data path
        data_path = "data"
        mock_variable_get.return_value = data_path

        with CliRunner().isolated_filesystem():
            release = UnpaywallRelease(self.unpaywall_test_file, self.unpaywall_test_date, self.unpaywall_test_date)
            # 'download' release
            shutil.copyfile(self.unpaywall_test_path, release.filepath_download)

            decompress_file_path = extract_release(release)
            decompress_file_name = os.path.basename(decompress_file_path)

            self.assertTrue(os.path.exists(decompress_file_path))
            self.assertEqual(self.unpaywall_test_decompress_file_name, decompress_file_name)
            self.assertEqual(self.unpaywall_test_decompress_hash, _hash_file(decompress_file_path, algorithm="md5"))

    @patch("observatory.platform.utils.workflow_utils.Variable.get")
    def test_transform_release(self, mock_variable_get):
        """Test that the release is transformed as expected.

        :return: None.
        """

        # Create data path and mock getting data path
        data_path = "data"
        mock_variable_get.return_value = data_path

        with CliRunner().isolated_filesystem():
            release = UnpaywallRelease(self.unpaywall_test_file, self.unpaywall_test_date, self.unpaywall_test_date)
            shutil.copyfile(self.unpaywall_test_path, release.filepath_download)

            extract_release(release)
            transform_file_path = transform_release(release)
            transform_file_name = os.path.basename(transform_file_path)

            self.assertTrue(os.path.exists(transform_file_path))
            self.assertEqual(self.unpaywall_test_transform_file_name, transform_file_name)
            self.assertEqual(self.unpaywall_test_transform_hash, _hash_file(transform_file_path, algorithm="md5"))



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, James Diprose

"""
A DAG that harvests the Unpaywall database: https://unpaywall.org/

Saved to the BigQuery table: <project_id>.our_research.unpaywallYYYYMMDD

Has been tested with the following Unpaywall releases:
* 2020-04-27, 2020-02-25, 2019-11-22, 2019-08-16, 2019-04-19, 2019-02-21, 2018-09-27, 2018-09-24

Does not work with the following releases:
* 2018-03-29, 2018-04-28, 2018-06-21, 2018-09-02, 2018-09-06
"""

import pendulum
from airflow import DAG
from airflow.operators.python import PythonOperator, ShortCircuitOperator
from academic_observatory_workflows.workflows.unpaywall_telescope import UnpaywallTelescope

default_args = {"owner": "airflow", "start_date": pendulum.datetime(2018, 9, 7)}

with DAG(dag_id="unpaywall", schedule_interval="@weekly", default_args=default_args) as dag:
    # Check that variables exist
    check = PythonOperator(
        task_id=UnpaywallTelescope.TASK_ID_CHECK_DEPENDENCIES,
        python_callable=UnpaywallTelescope.check_dependencies,
        queue=UnpaywallTelescope.QUEUE,
    )

    # List releases and skip all subsequent tasks if there is no release to process
    list_releases = ShortCircuitOperator(
        task_id=UnpaywallTelescope.TASK_ID_LIST,
        python_callable=UnpaywallTelescope.list_releases,
        queue=UnpaywallTelescope.QUEUE,
    )

    # Downloads snapshot from url
    download = PythonOperator(
        task_id=UnpaywallTelescope.TASK_ID_DOWNLOAD,
        python_callable=UnpaywallTelescope.download,
        queue=UnpaywallTelescope.QUEUE,
        retries=UnpaywallTelescope.RETRIES,
    )

    # Upload downloaded files for safekeeping
    upload_downloaded = PythonOperator(
        task_id=UnpaywallTelescope.TASK_ID_UPLOAD_DOWNLOADED,
        python_callable=UnpaywallTelescope.upload_downloaded,
        queue=UnpaywallTelescope.QUEUE,
        retries=UnpaywallTelescope.RETRIES,
    )

    # Decompresses download
    extract = PythonOperator(
        task_id=UnpaywallTelescope.TASK_ID_EXTRACT,
        python_callable=UnpaywallTelescope.extract,
        queue=UnpaywallTelescope.QUEUE,
    )

    # Transforms download
    transform = PythonOperator(
        task_id=UnpaywallTelescope.TASK_ID_TRANSFORM,
        python_callable=UnpaywallTelescope.transform,
        queue=UnpaywallTelescope.QUEUE,
    )

    # Upload download to gcs bucket
    upload_transformed = PythonOperator(
        task_id=UnpaywallTelescope.TASK_ID_UPLOAD_TRANSFORMED,
        python_callable=UnpaywallTelescope.upload_transformed,
        queue=UnpaywallTelescope.QUEUE,
        retries=UnpaywallTelescope.RETRIES,
    )

    # Upload download to BigQuery table
    bq_load = PythonOperator(
        task_id=UnpaywallTelescope.TASK_ID_BQ_LOAD,
        python_callable=UnpaywallTelescope.load_to_bq,
        queue=UnpaywallTelescope.QUEUE,
    )

    # Delete locally stored files
    cleanup = PythonOperator(
        task_id=UnpaywallTelescope.TASK_ID_CLEANUP,
        python_callable=UnpaywallTelescope.cleanup,
        queue=UnpaywallTelescope.QUEUE,
    )

    (
        check
        >> list_releases
        >> download
        >> upload_downloaded
        >> extract
        >> transform
        >> upload_transformed
        >> bq_load
        >> cleanup
    )
